# -*- coding: utf-8 -*-
"""
Injector: A Transition class that allows a injection of new members to a populations at
a specific point in the evolution

@author:karlen
"""

from pypmca.Parameter import Parameter
from pypmca.Population import Population
from pypmca.Transition import Transition


class Injector(Transition):
    """
    An injection object will cause the number in a population to change at specified point
    in the evolution.

    The time_step for which the action is taken can be specified as one of 'time_spec':
    - 'rel_days': int, number of days after start (t0)
    - 'rel_steps': int, number of time_steps after start (t0)

    - to_population: will have its numbers increased by an amount defined by
    - injection: a float type Parameter object

    In all cases, the time is specified by a Parameter object

    """

    def __init__(self, transition_name: str, time_spec: str, transition_time: Parameter,
                 to_population: Population, injection: Parameter, enabled: bool = True, model=None):
        """Constructor
        """
        super().__init__(transition_name, time_spec, transition_time, enabled, model)

        if not isinstance(to_population, Population):
            raise TypeError('Injector (' + self.name +
                            ') to_population must be a Population object')
        self.to_population = to_population

        if injection is None:
            raise TypeError('Error in constructing transition (' + self.name +
                            '): injection cannot be None')

        if not isinstance(injection, Parameter):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): injection must be a Parameter object')

        if injection.parameter_type != 'float':
            raise TypeError('Error in constructing transition (' + self.name +
                            '): injection must be a float type Parameter object')
        self.injection = injection

        self.parameters[str(injection)] = injection

    def take_action(self, expectations=True):
        """
        Inject the population
        """
        value = self.injection.get_value()
        if expectations:
            self.to_population.update_future_fast(value)
        else:
            self.to_population.update_future_fast(int(round(value)))

    def reset(self):
        """
        Nothing to reset
        """



# -*- coding: utf-8 -*-
"""
Chain: A Connector class processes a chain of one-to-one propagators, and
accumulating the remainder for a new population

@author: karlen
"""
from scipy import stats
import numpy as np

from pypmca.Connector import Connector
from pypmca.Delay import Delay
from pypmca.Parameter import Parameter
from pypmca.Population import Population
from pypmca.Propagator import Propagator


class Chain(Connector):
    """
    Chain: processes multiple one-to-one propagators and
    sends the remainder of the population to the to_population
        - connector_name: string, short descriptor
        - from_population: Population object at the start of the chain
        - to_population: Population object that receives the remainder
        the destination populations.
        - chain: list of one-to-one Propagator objects with 'norm' delays
        - fraction: Parameter object with expected
        fraction of remaining population to be propagated to the to_population,
        - delay: Delay object that defines how the remainder
        propagation is spread over time.
    """

    def __init__(self, connector_name: str, from_population: Population, to_population: Population,
                 chain: list, fraction: Parameter, delay: Parameter, model):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, Population):
            raise TypeError('Chain (' + self.name +
                            ') from_population must be a Population object')

        if not isinstance(to_population, Population):
            raise TypeError('Chain (' + self.name +
                            ') to_population must be a Population object')

        if not isinstance(chain, list):
            raise TypeError('Chain (' + self.name +
                            ') chain must be a list of Propagator objects')

        for prop in chain:
            if not isinstance(prop, Propagator):
                raise TypeError('Chain (' + self.name +
                                ') chain must be a list of only Propagator objects')
            if not isinstance(prop.from_population, Population):
                raise ValueError('Chain (' + self.name +
                                 ') chain must be a list of one-to-one Propagator objects')
            if not isinstance(prop.to_population, Population):
                raise ValueError('Chain (' + self.name +
                                 ') chain must be a list of one-to-one Propagator objects')
            if prop.delay.delay_type != 'norm':
                raise ValueError('Chain (' + self.name +
                                 ') chain must be a list of Propagator objects with "norm" delays')

            for key in prop.delay.delay_parameters:
                # notify parameter that this object's update method must be call if changed
                prop.delay.delay_parameters[key].set_must_update(self)

            prop.fraction.set_must_update(self)

        self.chain = chain

        if not isinstance(fraction, Parameter):
            raise TypeError('Chain (' + self.name +
                            ') fraction must be a Parameter object')
        if fraction.get_value() < 0.:
            raise ValueError('Chain (' + self.name +
                             ') fraction cannot be negative')
        if fraction.get_value() > 1.:
            raise ValueError('Chain (' + self.name +
                             ') fraction cannot exceed 1')
        self.fraction = fraction

        if not isinstance(delay, Delay):
            raise TypeError('Chain (' + self.name +
                            ') delay must be a Delay object')
        self.delay = delay

        if model is None:
            raise TypeError('Chain (' + self.name + ') model cannot be None')
        if not hasattr(model, 'get_time_step'):
            raise TypeError('Chain (' + self.name +
                            ') model must be a model object')
        self.model = model

        self.parameters['fraction'] = self.fraction
        if delay.delay_parameters is not None:
            for key in delay.delay_parameters:
                self.parameters['delay_' + key] = delay.delay_parameters[key]

        self.propagators = None
        self.remainder_propagator = None
        self.__setup_compound_propagation()

    # the same issue for delay parameters (update required after a change)
    def update(self):
        self.__setup_compound_propagation()

    def __setup_compound_propagation(self):
        """ For each link in chain, work out its effective combined delay
        and the combined fraction, to define a set of propagators """
        remainder_fraction = 0.
        current_fraction = 1.
        mean = 0.
        sigma = 0.
        self.propagators = []
        for prop in self.chain:
            prop_frac = prop.fraction.get_value()
            remainder_fraction += current_fraction * (1. - prop_frac)
            current_fraction *= prop_frac
            mean += prop.delay.delay_parameters['mean'].get_value()
            prop_sig = prop.delay.delay_parameters['sigma'].get_value()
            sigma2 = sigma * sigma + prop_sig * prop_sig
            sigma = np.sqrt(sigma2)

            frac = Parameter('frac', prop_frac)
            delay_pars = {
                'mean': Parameter('mean', mean, parameter_min=0., parameter_max=2.*mean),
                'sigma': Parameter('sigma', sigma, parameter_min= 0., parameter_max=2.*mean)
            }
            delay = Delay('delay_'+str(prop), 'norm', delay_pars, self.model)
            self.propagators.append(
                Propagator(prop.name + '_chain', self.from_population,
                           prop.to_population, frac, delay))
        frac = Parameter('frac', remainder_fraction * self.fraction.get_value())
        self.remainder_propagator = \
            Propagator(self.name + '_remainder', self.from_population,
                       self.to_population, frac, self.delay)

    def update_expectation(self):
        """
        Calculate contributions to other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]

            current_fraction = 1.
            for prop in self.propagators:
                current_fraction *= prop.fraction.get_value()
                scale = current_fraction * incoming
                prop.to_population.update_future_expectation(scale, prop.delay)

            prop = self.remainder_propagator
            scale = prop.fraction.get_value() * incoming
            prop.to_population.update_future_expectation(scale, prop.delay)

    def update_data(self):
        """
        Simulate data for other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]

            remaining = 0
            in_chain = incoming
            for prop in self.propagators:
                frac = prop.fraction.get_value()
                selected = stats.binom.rvs(in_chain, frac)
                remaining += (in_chain - selected)
                in_chain = selected
                prop.to_population.update_future_data(selected, prop.delay)

            prop = self.remainder_propagator
            # note that there is no need to apply the remainder fraction here
            selected = stats.binom.rvs(remaining, self.fraction.get_value())
            prop.to_population.update_future_data(selected, prop.delay)



# -*- coding: utf-8 -*-
"""
Transition: A base'virtual' class for defining transitions during the evolution

The derived classes must define the action methods


@author:karlen
"""

from pypmca.Parameter import Parameter


class Transition:
    """
    Base class for defining transitions during the evolution

    The time_step for which the action is taken can be specified as one of 'time_spec':
    - 'rel_days': float, number of days after start (t0)
    - 'rel_steps': int, number of time_steps after start (t0)
    
    The transition can be disabled by setting enabled=False
    
    enabled: specifies default state for the transition

    In all cases, the time is specified by a Parameter object
    """

    TIME_SPECS = {'rel_days': 'int', 'rel_steps': 'int'}

    def __init__(self, transition_name, time_spec, transition_time,
                 enabled=True, model=None):
        """Constructor
        """
        self.name = str(transition_name)

        if not isinstance(time_spec, str):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): time_spec argument must be a str')

        if time_spec not in self.TIME_SPECS:
            specs = []
            for key in self.TIME_SPECS:
                specs.append(key)
            buff = '/'.join(specs)
            raise TypeError('Error in constructing transition (' + self.name +
                            '): time_spec argument must be one of' + buff)
        self.time_spec = time_spec

        if not isinstance(transition_time, Parameter):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): transition_time argument must be a Parameter object')

        if transition_time.parameter_type != self.TIME_SPECS[self.time_spec]:
            raise TypeError('Transition (' + self.name +
                            ') transition time (' +
                            transition_time.parameter_type +
                            ') does not match time_spec (' +
                            self.time_spec + ')')

        if transition_time.get_value() <= 0:
            raise ValueError('Transition (' + self.name +
                             ') transition time must be larger than zero.')

        self.transition_time = transition_time
        self.transition_time.set_must_update(self)

        if model is None:
            raise TypeError('Transition (' + self.name + ') model cannot be None')
        if not hasattr(model, 'get_time_step'):
            raise TypeError('Transition (' + self.name +
                            ') model must be a model object')
        self.model = model

        self.trigger_step = None
        self.__calculate_trigger_step()

        self.parameters = {str(self.transition_time): self.transition_time}

        self.enabled = enabled

    def update(self):
        """ Method called if transition_time parameter changed
        """
        self.__calculate_trigger_step()

    def __calculate_trigger_step(self):
        """ calculate the step when this transition should take place
        """
        time_step = self.model.get_time_step()
        day = 1. / time_step

        if self.time_spec == 'rel_days':
            self.trigger_step = int(round(self.transition_time.get_value() * day))
        elif self.time_spec == 'rel_steps':
            self.trigger_step = self.transition_time.get_value()

    def __str__(self):
        return self.name

    def take_action(self, expectations=True):
        """
        This method must be defined in derived classes

        Raises
        ------
        NotImplementedError if not defined in the derived class

        Returns
        -------
        None.

        """
        raise NotImplementedError()

    def reset(self):
        """
        This method must be defined in derived classes

        Raises
        ------
        NotImplementedError if not defined in the derived class

        Returns
        -------
        None.

        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
Ensemble: A Model class whose behaviour is defined by an ensemble of models

An ensemble allows for categorization of populations by age, risk, or other factors. It can also
be used to combine many models that may or may not benearly independent
(such as separate provinces to make a Canada wide model).

Each category has its own model, and one model can influence the growth within other models.
The ensemble sums the histories of all its models to represent the evolution of
the entire system.

When mixing models having different growth rates, achieving the desired initial condition at t_0 through the
boot process is challenging. The boot process starts with a small number in each model's boot_population and
the boot ends when the ensemble exceeds its goal. It is possible that the relative sizes for each model
would differ significantly from the desired proportions, at t=0, and after scaling each model for the t_0 condition,
the state would be far from a steady state. A second boot can be done (by adjusting the small numbers in each model's
boot_population) according to the outcome from the first boot - ie. doing an iterative boot.
This issue is not as serious if, at least initially, the growth behaviours of the different groups are similar.

Owing to the complexity of ensembles made from mixtures of very different growth rate sub groups,
such situations should be treated with care until such behaviour has been thoroughly tested!

@author: karlen
"""
from datetime import date
import numpy as np
from scipy import stats
import copy

import pickle
from pypmca import Model, Parameter, Population


class Ensemble(Model):
    """
    Ensemble overwrites some model methods, so that actions are applied to all models.
    It's own connectors are not used for evolution: each model is evolved and ensemble
    histories are just the sum of the models histories.
    """

    def __init__(self, name: str, model):
        """Constructor
         - Model specifies the reference model: populations etc should match those in
         the ensemble of models. It can be specified as a model, or as a filename
        """
        super().__init__(name)

        self.model = None
        in_model = None
        if isinstance(model, Model):
            in_model = model
        elif isinstance(model, str):
            in_model = Model.open_file(model)
        else:
            raise TypeError('Error creating ensemble. The model argument needs to be a model object ' +
                            'or have a .pypm filename')

        # to avoid the possibility that the reference model is subsequently modified, make a copy
        # that will not be as exposed to tampering
        self.model = copy.deepcopy(in_model)

        # point to the population objects in self.model to save the histories
        # this allows for consistent access to history between models and ensembles
        for pop_name in self.model.populations:
            pop = self.model.populations[pop_name]
            self.populations[pop_name] = pop

        # the parameters will be set once the list of models are read in
        self.parameters = {}

        # the boot parameters point to the ones copied from the reference
        self.boot_pars = self.model.boot_pars

        # the important scaling parameter for the ensemble gets defined here
        # this is forced to be a parameter when each model is created and
        # checked when the models are included in the ensemble

        boot_pop = self.populations[self.boot_pars['boot_population']]
        boot_pop_ip = boot_pop.initial_value
        self.parameters[boot_pop_ip.name] = boot_pop_ip
        # When this scaling parameter is changed, all the model scaling parameters
        # must also be scaled accordingly.
        boot_pop_ip.set_must_update(self)

        # The list of models (either provided directly or as a list of .pypm files)
        # They are required to have unique names for user interaction
        # Ordering is preserved to match the transmission matrix
        self.model_list = []
        self.models = {}
        self.infected_name = None
        self.susceptible_name = None
        self.total_name = None
        self.contagious_name = None
        self.alpha_name = None
        self.infection_cycle_name = None
        self.contact = None
        self.contact_matrix = None
        self.contact_type = ''
        self.null_pop = Population('null', 0.)
        self.total_goal = None
        self.boot_achieved = {}
        self.max_scale_factor = None

        self.__distribution = None
        self.__nbinom_par = None
        self.set_distribution('poisson', None)

    def set_distribution(self, distribution, nbinom_par):
        if distribution not in ['poisson', 'nbinom']:
            raise ValueError('Ensemble (' + self.name +
                             ') distribution must be poisson or nbinom')
        self.__distribution = distribution

        if distribution == 'nbinom' and \
                (nbinom_par is None or not isinstance(nbinom_par, Parameter)):
            raise TypeError('Ensemble (' + self.name +
                            ') nbinom_par must be a Parameter object')
        self.__nbinom_par = nbinom_par

    def get_distribution(self):
        return self.__distribution, self.__nbinom_par

    def update(self):
        """
        This is called when the ensemble scaling parameter is changed. It is necessary
        to update the model scaling parameters accordingly.
        """
        ens_scale = 0.
        for model_name in self.models:
            model = self.models[model_name]
            boot_pop = model.populations[model.boot_pars['boot_population']]
            ens_scale += boot_pop.initial_value.get_value()
        if ens_scale > 0.:
            ens_boot_pop = self.populations[self.boot_pars['boot_population']]
            ratio = ens_boot_pop.initial_value.get_value()/ens_scale
            for model_name in self.models:
                model = self.models[model_name]
                boot_pop = model.populations[model.boot_pars['boot_population']]
                new_scale = boot_pop.initial_value.get_value() * ratio
                boot_pop.initial_value.set_value(new_scale)

            self.boot_needed = True

    def upload_models(self, models: list):
        if not isinstance(models, list):
            raise TypeError('Error in upload_models: argument must be a list of Model objects')
        for model in models:
            if not isinstance(model, Model):
                raise TypeError('Error in upload_models: argument must be a list of Model objects')
        for model in models:
            self.__add_model(model)

    def upload_model_files(self, model_filenames: list):
        if not isinstance(model_filenames, list):
            raise TypeError('Error in upload_model_files: argument must be a list of .pypm filenames')
        for model_file in model_filenames:
            if not isinstance(model_file, str):
                raise TypeError('Error in upload_model_files: argument must be a list of .pypm filenames')

        for model_filename in model_filenames:
            self.__add_model(Model.open_file(model_filename))

    def __add_model(self, model):
        if model.name in self.models:
            raise ValueError('Error: Two models have same name: ' + model.name)
        for population_name in self.populations:
            if population_name not in model.populations:
                raise ValueError('Error: Model ' + model.name + ' does not contain population:' + population_name)
        if model.boot_pars['boot_population'] != self.model.boot_pars['boot_population']:
            raise ValueError('Error: Model ' + model.name + ' boot population name (' + model.boot_pars['boot_population'] +
            ') does not match name in reference boot population (' + self.model.boot_pars['boot_population'])
        boot_pop_ip = model.populations[model.boot_pars['boot_population']].initial_value
        if not isinstance(boot_pop_ip, Parameter):
            raise ValueError(
                'Error: Model ' + model.name + ' boot population (' + model.boot_pars['boot_population'] +
                ') is not initialized by a parameter')

        self.model_list.append(model.name)
        self.models[model.name] = model
        self.reset_cross_transmission()

        # include parameters
        model_app = '#' + str(len(self.model_list) - 1)
        for par_name in model.parameters:
            new_name = par_name + model_app
            self.parameters[new_name] = model.parameters[par_name]

    def define_cross_transmission(self, infection_cycle_name: str, infected_name: str,
                                  susceptible_name: str, total_name: str,
                                  contagious_name: str, alpha_name: str,
                                  contact_type: str = 'independent', contact: list = None):
        """
        Define how the populations mix to produce new infections:
        For example new infections in model A arising from interactions with group B
        are calculated by Susceptible_A / M * f[A][B] * Contagious_B * alpha_AB
            - M is the total effective total population = sum f[A][B] * N_B.
            - alpha_AB is the geometric mean = sqrt(alpha_A * alpha_B)
            - the contact matrix, f, needs to be n x n, where n is the number of models
            and the order of the matrix rows matches the order of the models in model_list
        In a homogeneous society, all alphas are the same, and all terms of the matrix are 1.
        f represents the relative probability for a contagious members of B group to infect a
        random member of the A group (relative to infecting a random member of the B group).
        The off diagonal elements are typically less than 1. The diagonal elements are 1 by
        definition. A diagonal matrix describes a set of independent populations.
        Given the definintion of M, both homogenous and independent populations yield their
        original infection rate without adjusting alpha.

        The contact matrix can be specified in a number of ways, defined by contact_type.
            - 'diagonal','independent': all diagonal terms are 1, non-diagonal are 0.
            This corresponds to the situation where all models develop independently
            - 'equality': all matrix elements are 1.
            This corresponds to the situation where members are blind to the catagorizations
            - 'fixed': an arbitrary contact_matrix is provided as a list of list of floats
            Diagonal elements are inforced to be 1. An ValueError is raised if not.
            - 'simple': all off-diagonal elements are equal and specified by a single
            Parameter object provided in list passed by the contact argument
            - 'symmetric': off-diagonal elements passed by n(n-1)/2 Parameter objects
            provided in the list passed by the contact argument - f12,f13,...,f1n,f23,f24...
        Parameters are included in the ensemble parameter list for ease in adjusting the mixing.

        """
        if infection_cycle_name not in self.model.connectors:
            raise ValueError('Infection cycle (' + infection_cycle_name + ') not found in connectors')
        self.infection_cycle_name = infection_cycle_name
        # these could have been derived from belpw, but for non-symmetric contact matricies
        # the order of the two multipliers matter.
        if infected_name not in self.populations:
            raise ValueError('Population ' + infected_name + ' not found in reference model.')
        self.infected_name = infected_name
        if susceptible_name not in self.populations:
            raise ValueError('Population ' + susceptible_name + ' not found in reference model.')
        self.susceptible_name = susceptible_name
        if total_name not in self.populations:
            raise ValueError('Population ' + total_name + ' not found in reference model.')
        self.total_name = total_name
        if contagious_name not in self.populations:
            raise ValueError('Population ' + contagious_name + ' not found in reference model.')
        self.contagious_name = contagious_name
        if alpha_name not in self.model.parameters:
            raise ValueError('Parameter ' + alpha_name + ' not found in reference model.')
        self.alpha_name = alpha_name

        self.contact_type = contact_type
        n_model = len(self.model_list)
        if contact_type == 'fixed':
            if not isinstance(contact, list):
                raise TypeError('The contact matrix must be a list of lists')
            if len(contact) != n_model:
                raise ValueError('The contact matrix must be an nxn matrix, where n=number of models')
            for i in range(n_model):
                row = contact[i]
                if not isinstance(row, list):
                    raise TypeError('The contact matrix must be a list of lists')
                if len(row) != n_model:
                    raise ValueError('The contact matrix must be an nxn matrix, where n=number of models')
                if np.abs(contact[i][i] - 1.) > 0.001:
                    raise ValueError('The diagonal terms in the contact matrix should be 1.')
            self.contact_matrix = contact
        elif contact_type == 'simple':
            if not isinstance(contact, list):
                raise TypeError('The contact argument must be a list of Parameters')
            if len(contact) != 1:
                raise ValueError('The contact argument must be a list of length 1, ')
            for cont in contact:
                if not isinstance(cont, Parameter):
                    raise TypeError('The contact argument must be a list containing a Parameter object')
        elif contact_type == 'symmetric':
            if not isinstance(contact, list):
                raise TypeError('The contact argument must be a list of Parameters')
            if len(contact) != n_model * (n_model - 1) / 2:
                raise ValueError('The contact argument must be a list with ' +
                                 str(n_model * (n_model - 1) / 2) + ' parameters')
            for cont in contact:
                if not isinstance(cont, Parameter):
                    raise TypeError('The contact argument must be a list of Parameter objects')
        elif contact_type not in ['diagonal', 'independent', 'equality']:
            raise ValueError('Contact matrix type not recognized.')

    def reset_cross_transmission(self):
        self.infected_name = None
        self.susceptible_name = None
        self.total_name = None
        self.contagious_name = None
        self.alpha_name = None
        self.contact_matrix = None
        self.contact_type = ''

    def set_t0(self, year=2020, month=1, day=1):
        """Define the date of the first element in the history lists
        """
        self.t0 = date(year, month, day)
        for model_name in self.models:
            model = self.models[model_name]
            model.set_t0(year, month, day)

    def set_time_step(self, time_step):

        if not isinstance(time_step, float):
            raise TypeError('Model (' + self.name +
                            ' time_step must be a float')
        if not time_step > 0.:
            raise ValueError('Model (' + self.name +
                             ' time_step must be positive')
        self.__time_step = time_step

    def get_time_step(self):
        """
        Get the length of time corresponding to one step in the evolution

        Returns
        -------
        real value for one time step

        """
        return self.__time_step

    def reset(self):
        """
        Reinitialize all populations, eliminating history and future, and
        setting initial values. Also reset transitioned parameters
        """
        for model_name in self.models:
            model = self.models[model_name]
            model.reset()
        self.__reset()

    def __reset(self):
        """
        Reinitialize all populations, eliminating history and future, and
        setting initial values. Also reset transitioned parameters
        """
        for key in self.populations:
            self.populations[key].reset()

        self.boot_needed = True

    def do_boot(self, expectations=True):
        if self.contact_type in ['diagonal', 'independent']:
            # in this case, boot each model separately
            # each will boot to its own goal

            # turn on the infection cycle in each model, in case it was disabled previously
            self.__enable_infections()

            # start from a clean slate of populations set to their initial values
            self.__reset()

            # need to set this to False before evolving expectations
            self.boot_needed = False

            for model_name in self.models:
                model = self.models[model_name]
                model.boot(expectations)

            # prior to full calculation, setup contact matrix etc.

            # turn off the infection cycle in each model, as the ensemble does this.
            self.__disable_infections()

            # setup the contact matrix
            self.__setup_contact_matrix()

        else:

            # try booting once, then a second time in case the scaling factors are too large
            self.boot_achieved = {}
            self.boot(expectations)

            # second boot in case scale factors are large
            # always do this to avoid discontinuities
            self.boot(expectations)

    def boot(self, expectations=True):
        """ Using information in boot_pars, evolve the system to a point where
        the target population in the ensemble is reached or exceeded:
            - expectations: if True, after the boot completes, the
            history will begin with an expectation,
            otherwise, a Poisson number is drawn using that expectation.
        Note: check is in place so that if boot requirement is not made within
        10k steps, it fails (rather than stay in infinite loop)

        The boot process itself may need to be iterated. If the relative sizes of the
        target population differs by a lot from the intended values, then the scaled
        populations after the boot may not be consistent with a steady state solution.
        """
        if len(self.boot_pars) == 0:
            raise RuntimeError('Model boot parameters not set.')

        # start from a clean slate of populations set to their initial values
        self.reset()

        # need to set this to False before evolving expectations
        self.boot_needed = False

        # save the goal_values for each model
        goal_values = {}
        goal_value = 0
        boot_value = 0
        for model_name in self.models:
            model = self.models[model_name]
            model_boot_pop = model.populations[model.boot_pars['boot_population']]
            goal_values[model_name] = model_boot_pop.history[-1]
            goal_value += model_boot_pop.history[-1]
            boot_value += model.boot_pars['boot_value']

        # the ensemble starts with a small population size in the boot_pop (defined by reference model)
        # the goal is to have this population grow to the sum of the sizes in each model
        boot_pop = self.populations[self.boot_pars['boot_population']]
        boot_pop.history[-1] = self.boot_pars['boot_value']
        # this small number needs to be distributed accordingly to the models
        # If this is the first boot, the best guess is to distribute the small number in proportion to
        # the goal numbers for each model. On a subsequent iteration, the scaling from the
        # results of the previous boot will provide a better solution.
        if len(self.boot_achieved) == 0:
            # First boot:
            self.total_goal = 0.
            for model_name in self.models:
                model = self.models[model_name]
                model_boot_pop = model.populations[model.boot_pars['boot_population']]
                self.total_goal += model_boot_pop.history[-1]
            if self.total_goal == 0.:
                raise ValueError('Targets in models are zero. Unable to boot.')
            boot_scale = 1.*boot_value/self.total_goal
            for model_name in self.models:
                model = self.models[model_name]
                model_boot_pop = model.populations[model.boot_pars['boot_population']]
                model_goal = model_boot_pop.history[-1]
                model_boot_pop.history[-1] = model_goal * boot_scale
        else:
            # Second boot:
            # the first boot derived fractions for each model based on goal/total
            # eg. desired population 10 and 30, the fractions are 0.25 and 0.75.
            # If that achieved relative populations after boot significantly different
            # eg.  2 and 38, then the starting fractions need to be adjusted to get
            # closer to the goal propotions after boot.
            # The original proportioning (goal/total) is revised by scaling each by goal/achieved
            # The new scaling is therefore goal^2/(achieved*total). If achieved is close to goal
            # this is a small change to the proportioning.
            total_rev = 0.
            for model_name in self.models:
                model = self.models[model_name]
                model_boot_pop = model.populations[model.boot_pars['boot_population']]
                total_rev += (model_boot_pop.history[-1])**2/self.total_goal/self.boot_achieved[model_name]
            boot_scale = 1. * boot_value / total_rev
            for model_name in self.models:
                model = self.models[model_name]
                model_boot_pop = model.populations[model.boot_pars['boot_population']]
                model_goal = model_boot_pop.history[-1]
                model_boot_pop.history[-1] = model_goal**2/self.total_goal/self.boot_achieved[model_name] * boot_scale

        # exclusion_populations for each model
        # are reset to their original values after the
        # boot. Save them here.

        model_exclusions = {}
        for model_name in self.models:
            model = self.models[model_name]
            exclusions = model.boot_pars['exclusion_populations']
            exc_vals = {}
            for exc_pop_name in exclusions:
                exc_pop = self.populations[exc_pop_name]
                exc_vals[exc_pop_name] = exc_pop.history[-1]
            model_exclusions[model_name] = exc_vals

        # turn off the infection cycle in each model, as the ensemble does this.
        self.__disable_infections()

        # setup the contact matrix
        self.__setup_contact_matrix()

        last_value = -1
        i_step = 0
        while boot_pop.history[-1] < goal_value and boot_pop.history[-1] >= last_value:
            last_value = boot_pop.history[-1]
            self.evolve_expectations(1)
            i_step += 1
            if i_step % 1000 == 0:
                if boot_pop.history[-1] < 0.1 * goal_value:
                    self.boot_needed = True
                    raise ValueError('The boot process is taking too long to complete. ' +
                                     'Adjust parameters to allow the goal to be reached.')

        if boot_pop.history[-1] < goal_value:
            raise ValueError('The boot process did not reach its target. Consider reducing the target or adjusting a parameter.')

        # save the achieved values: needed if there is a second boot
        # quality of boot is good if the scale factors are near 1
        self.max_scale_factor = 0.
        for model_name in self.models:
            model = self.models[model_name]
            model_boot_pop = model.populations[model.boot_pars['boot_population']]
            self.boot_achieved[model_name] = model_boot_pop.history[-1]
            scale_factor = model_boot_pop.history[-1]/goal_values[model_name]
            if scale_factor < 1.:
                scale_factor = 1./scale_factor
            if scale_factor > self.max_scale_factor:
                self.max_scale_factor = scale_factor

        ens_scale = goal_value / boot_pop.history[-1]

        # Erase models history until current. Scale all histories and futures.
        for model_name in self.models:
            model =self.models[model_name]
            exc_vals = model_exclusions[model_name]
            model_boot_pop = model.populations[model.boot_pars['boot_population']]
            scale = goal_values[model_name] / model_boot_pop.history[-1]
            for key in model.populations:
                pop = model.populations[key]
                pop.remove_history()
                if str(pop) not in exc_vals:
                    pop.scale_history(scale, expectations)
                else:
                    pop.history[-1] = exc_vals[str(pop)]
                pop.scale_future(scale, expectations)

        # restore goal_value for data production
        if not expectations:
            for model_name in self.models:
                model = self.models[model_name]
                model_boot_pop = model.populations[model.boot_pars['boot_population']]
                model_boot_pop.history[-1] = int(goal_values[model_name])

    def evolve_expectations(self, n_step):
        """
        First check if a boot is needed: this is required if
        reset_populations is called

        Take n_steps of
        0) check if it is time to apply a transition in any of the models
        1) updating internal futures in all the models
        2) update future due to infection from other models
        3) appending new value to history and remove current time step in future array
        """
        if self.boot_needed:
            self.do_boot(expectations=True)

        for step in range(n_step):
            self.do_transitions(step, expectations=False)

            # calculate influence between models
            self.__cross_model_transmission(expectations=True)

            # calculate future expectations
            self.calculate_future(expectations=True)

            # make one time step
            self.do_time_step(expectations=True)

        # add up the histories and put in local population objects
        self.__combine_histories()

    def generate_data(self, n_step):
        """
        Produce data
        """
        if self.boot_needed:
            self.do_boot(expectations=False)

        for step in range(n_step):
            self.do_transitions(step, expectations=False)

            # calculate influence between models
            self.__cross_model_transmission(expectations=False)

            # calculate future data
            self.calculate_future(expectations=False)

            # make one time step
            self.do_time_step(expectations=False)

        # add up the histories and put in local population objects
        self.__combine_histories()

    def do_transitions(self, step, expectations=True):
        for model_name in self.models:
            model = self.models[model_name]
            model.do_transitions(step, expectations)

    def calculate_future(self, expectations=True):
        for model_name in self.models:
            model = self.models[model_name]
            model.calculate_future(expectations)

    def __disable_infections(self):
        for model_name in self.models:
            model = self.models[model_name]
            model.connectors[self.infection_cycle_name].to_population = self.null_pop

    def __enable_infections(self):
        for model_name in self.models:
            model = self.models[model_name]
            model.connectors[self.infection_cycle_name].to_population = \
                model.populations[self.infected_name]

    def __setup_contact_matrix(self):
        n_model = len(self.model_list)
        if self.contact_type in ['diagonal', 'independent']:
            self.contact_matrix = []
            for i in range(n_model):
                row = [0.] * n_model
                row[i] = 1.
                self.contact_matrix.append(row)
        elif self.contact_type == 'equality':
            self.contact_matrix = []
            for i in range(n_model):
                row = [1.] * n_model
                self.contact_matrix.append(row)
        elif self.contact_type == 'simple':
            self.contact_matrix = []
            off_diag = self.contact.get_value()
            for i in range(n_model):
                row = [off_diag] * n_model
                row[i] = 1.
                self.contact_matrix.append(row)
        elif self.contact_type == 'symmetric':
            self.contact_matrix = [[0.] * n_model] * n_model
            i_pnt = 0
            for i in range(n_model):
                self.contact_matrix[i][i] = 1.
                for j in (i + 1, n_model):
                    self.contact_matrix[i][j] = self.contact[i_pnt].get_value()
                    self.contact_matrix[i][j] = self.contact[i_pnt].get_value()
                    i_pnt += 1

    def __cross_model_transmission(self, expectations=True):
        # check that everything is ready
        if not self.contact_type == 'fixed' and self.contact_matrix is None:
            raise RuntimeError('The cross-transmission information has not been provided yet.')

        n = len(self.model_list)

        for ia in range(n):
            model_name_A = self.model_list[ia]
            model_A = self.models[model_name_A]
            sum_denom = 0.
            sum_numer = 0.
            if self.contact_type in ['diagonal', 'independent']:
                sum_denom += model_A.populations[self.total_name].history[-1]
                term1 = model_A.parameters[self.alpha_name].get_value()
                term2 = model_A.populations[self.susceptible_name].history[-1]
                term3 = model_A.populations[self.contagious_name].history[-1]
                sum_numer += term1 * term2 * term3
            else:
                for ib in range(n):
                    model_name_B = self.model_list[ib]
                    model_B = self.models[model_name_B]

                    sum_denom += self.contact_matrix[ia][ib] * model_B.populations[self.total_name].history[-1]
                    term1 = np.sqrt(model_A.parameters[self.alpha_name].get_value() *
                                    model_B.parameters[self.alpha_name].get_value())
                    term2 = model_A.populations[self.susceptible_name].history[-1]
                    term3 = model_B.populations[self.contagious_name].history[-1]
                    sum_numer += term1 * term2 * term3 * self.contact_matrix[ia][ib]
            new_infections = sum_numer / sum_denom

            if expectations:
                model_A.populations[self.infected_name].update_future_fast(new_infections)
            else:
                if self.__distribution == 'poisson':
                    n = stats.poisson.rvs(new_infections)
                    model_A.populations[self.infected_name].update_future_fast(n)
                else:
                    p = self.__nbinom_par.get_value()
                    if p < 0.001:
                        p = 0.001
                    if p > 0.999:
                        p = 0.999
                    r = new_infections * p / (1. - p)
                    n = stats.nbinom.rvs(r, p)
                    model_A.populations[self.infected_name].update_future_fast(n)

    def do_time_step(self, expectations=True):
        for model_name in self.models:
            model = self.models[model_name]
            model.do_time_step(expectations)

    def __combine_histories(self):
        # add the histories of all models together
        the_models = []
        for model_name in self.models:
            model = self.models[model_name]
            the_models.append(model)

        for pop_name in self.populations:
            ens_pop = self.populations[pop_name]
            ens_history = [0] * len(the_models[0].populations[pop_name].history)
            for the_model in the_models:
                mod_history = the_model.populations[pop_name].history
                for i in range(len(ens_history)):
                    ens_history[i] += mod_history[i]
            ens_pop.history = ens_history

    def add_transition(self, transition):
        raise NotImplementedError()

    def add_connector(self, connector, before_connector=None, after_connector=None):
        """
        Overwrite this method since it is important to not change the populations in the reference model
        """
        raise NotImplementedError()

    def remove_connector(self, connector):
        raise NotImplementedError()

    def update_lists(self):
        raise NotImplementedError()

    def save_file(self, filename):
        """
        Save a copy of the current ensemble to a file. The ensemble can be restored
        at a late time using Ensemble.open_file(filename).

        Before saving, the model futures are removed, and the initial_values
        are set to the current values for all parameters.

        If no extension is provided, the default extension, .pypm_e is added.

        Parameters
        ----------
        filename : str
            name of file to save model

        Returns
        -------
        None.

        """
        if not isinstance(filename, str):
            raise TypeError('Error saving file. ' +
                            ': filename argument must be a str')

        fullname = filename
        if '.' not in filename:
            fullname = filename + '.pypm_e'

        #       eliminate histories, undo transitioned parameter adjustments

        self.reset()

        #       set initial parameter values to their current values to define a new
        #       reference

        for model_name in self.models:
            model = self.models[model_name]
            for par_name in model.parameters:
                par = model.parameters[par_name]
                par.new_initial_value()

        with open(fullname, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def open_file(cls, filename):
        """
        Restore an ensemble that was saved to a file using Ensemble.save_file(filename)

        Parameters
        ----------
        filename : str
            name of existing model file to open

        Returns
        -------
        Ensemble
            The ensemble object saved in the file

        """
        if not isinstance(filename, str):
            raise TypeError('Error opening file. ' +
                            ': filename argument must be a str')

        fullname = filename
        if '.' not in filename:
            fullname = filename + '.pypm_e'

        with open(fullname, 'rb') as f:
            return pickle.load(f)



# -*- coding: utf-8 -*-
"""
Propagator: A Connector class that propagates a fraction of its incoming
population to others, either in the next time step or distributed in time.

Note that propagation is done independently to the other populations.
To divide the incoming population into different sub-populations (dependently),
use the Split class instead.

There is no benefit in having a list of to_populations, since it is equivalent
to having two independent propagators.

@author: karlen
"""
from scipy import stats

from pypmca.Connector import Connector
from pypmca.Delay import Delay
from pypmca.Parameter import Parameter


class Propagator(Connector):
    """
    Propagator sends a fraction of an incoming population to other populations,
    either in the next time step or distributed in time:
        - connector_name: string, short descriptor
        - from_population: Population object, source of population to be
        propagated
        - to_population: Population object or list of Population objects,
        destination population(s).
        - fraction: Parameter object or list of Parameter objects, expected
        fraction of population to be propagated. If list provided,
        its length must match to_population, and the fractions are applied
        independently. The sum of all fractions can exceed 1.
        - delay: Delay object or list of Delay objects that define how
        propagation is spread over time. If list provided, its length
        must match length of to_population.
    """

    def __init__(self, connector_name: str, from_population, to_population,
                 fraction, delay: Delay):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if isinstance(from_population, list):
            raise TypeError('Propagator (' + self.name +
                            ') from_population cannot be a list')

        if isinstance(fraction, list):
            if not isinstance(to_population, list):
                raise TypeError('Propagator (' + self.name +
                                ') fraction cannot be a list when' +
                                ' to_population is not a list')
            if len(fraction) != len(to_population):
                raise ValueError('Propagator (' + self.name +
                                 ') fraction list length does not match' +
                                 ' to_population list length')
            for frac in fraction:
                if not isinstance(frac, Parameter):
                    raise TypeError('Propagator (' + self.name +
                                    ') fraction list must contain Parameter objects')
        elif not isinstance(fraction, Parameter):
            raise TypeError('Propagator (' + self.name +
                            ') fraction must be a Parameter object')
        self.fraction = fraction

        if not isinstance(delay, Delay):
            raise TypeError('Propagator (' + self.name +
                            ') delay must be a Delay object')
        self.delay = delay

        if isinstance(fraction, list):
            for i in range(len(fraction)):
                name = 'fraction_' + str(to_population[i])
                self.parameters[name] = self.fraction[i]
        else:
            self.parameters['fraction'] = self.fraction
        if isinstance(delay, list):
            for i in range(len(delay)):
                name = 'delay_' + str(to_population[i])
                if delay[i].delay_parameters is not None:
                    for key in delay[i].delay_parameters:
                        self.parameters[name + '_' + key] = delay[i].delay_parameters[key]
        else:
            if delay.delay_parameters is not None:
                for key in delay.delay_parameters:
                    self.parameters['delay_' + key] = delay.delay_parameters[key]

    def update_expectation(self):
        """
        Calculate contributions to other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]
            if isinstance(self.to_population, list):
                if isinstance(self.fraction, list):
                    if isinstance(self.delay, list):
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            scale = self.fraction[i].get_value() * incoming
                            to_pop.update_future_expectation(scale, self.delay[i])
                    else:
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            scale = self.fraction[i].get_value() * incoming
                            to_pop.update_future_expectation(scale, self.delay)
                else:
                    scale = self.fraction.get_value() * incoming
                    if isinstance(self.delay, list):
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            to_pop.update_future_expectation(scale, self.delay[i])
                    else:
                        for to_pop in self.to_population:
                            to_pop.update_future_expectation(scale, self.delay)
            else:
                scale = self.fraction.get_value() * incoming
                self.to_population.update_future_expectation(scale, self.delay)

    def update_data(self):
        """
        Simulate data for other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]
            if isinstance(self.to_population, list):
                if isinstance(self.fraction, list):
                    if isinstance(self.delay, list):
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            scale = stats.binom.rvs(incoming, self.fraction[i].get_value())
                            to_pop.update_future_data(scale, self.delay[i])
                    else:
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            scale = stats.binom.rvs(incoming, self.fraction[i].get_value())
                            to_pop.update_future_data(scale, self.delay)
                else:
                    if isinstance(self.delay, list):
                        for i in range(len(self.to_population)):
                            to_pop = self.to_population[i]
                            scale = stats.binom.rvs(incoming, self.fraction.get_value())
                            to_pop.update_future_data(scale, self.delay[i])
                    else:
                        for to_pop in self.to_population:
                            scale = stats.binom.rvs(incoming, self.fraction.get_value())
                            to_pop.update_future_data(scale, self.delay)
            else:
                scale = stats.binom.rvs(incoming, self.fraction.get_value())
                self.to_population.update_future_data(scale, self.delay)



# -*- coding: utf-8 -*-
"""
Splitter: A Connector class that distributes its incoming
population to n others, either in the next time step or distributed in time.

@author: karlen
"""
from scipy import stats

from pypmca.Connector import Connector
from pypmca.Delay import Delay
from pypmca.Parameter import Parameter


class Splitter(Connector):
    """
    Splitter distributes its incoming population to other populations,
    either in the next time step or distributed in time:
        - connector_name: string, short descriptor
        - from_population: Population object, source of population to be
        propagated
        - to_population: list of two or more Population objects,
        the destination populations.
        - fractions: list of Parameter object with expected
        fraction of population to be propagated to the first to_population,
        the second is the fraction of the remaining to go to the next to_population
        and so on, with the remainder going to the other population
        The fractions must be in the range (0.,1.)
        - delay: Delay object or list of the Delay objects that define how
        propagation is spread over time.
    """

    def __init__(self, connector_name: str, from_population, to_population: list,
                 fractions: list, delay: Delay):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if isinstance(from_population, list):
            raise TypeError('Splitter (' + self.name +
                            ') from_population cannot be a list')

        if not isinstance(to_population, list):
            raise TypeError('Splitter (' + self.name +
                            ') to_population must be a list')

        if len(to_population) < 2:
            raise ValueError('Splitter (' + self.name +
                             ') to_population must be a list of length > 1')

        if not isinstance(fractions, list):
            raise TypeError('Splitter (' + self.name +
                            ') fractions must be a list object')

        if len(fractions) != len(to_population) - 1:
            raise ValueError('Splitter (' + self.name +
                             ') fractions length must be len(to_population)-1')

        for fraction in fractions:
            if not isinstance(fraction, Parameter):
                raise TypeError('Splitter (' + self.name +
                                ') fractions list must contain Parameter object')

            if fraction.get_value() < 0.:
                raise ValueError('Splitter (' + self.name +
                                 ') fraction cannot be negative')
            if fraction.get_value() > 1.:
                raise ValueError('Splitter (' + self.name +
                                 ') fraction cannot exceed 1')

        self.fractions = fractions

        if isinstance(delay, list):
            for d in delay:
                if not isinstance(d, Delay):
                    raise TypeError('Splitter (' + self.name +
                                    ') delay list must only have Delay objects')

            if len(delay) != len(to_population):
                raise ValueError('Splitter (' + self.name +
                                 ') delay length must match to_population length' +
                                 ' or be a single (common) delay')

        elif not isinstance(delay, Delay):
            raise TypeError('Splitter (' + self.name +
                            ') delay must be a Delay object')
        self.delay = delay

        for i in range(len(self.fractions)):
            self.parameters['fraction' + str(i)] = self.fractions[i]

        if isinstance(delay, list):
            for i in range(len(delay)):
                name = 'delay_' + str(to_population[i])
                if delay[i].delay_parameters is not None:
                    for key in delay[i].delay_parameters:
                        self.parameters[name + key] = delay[i].delay_parameters[key]
        else:
            if delay.delay_parameters is not None:
                for key in delay.delay_parameters:
                    self.parameters['delay_' + key] = delay.delay_parameters[key]

    def update_expectation(self):
        """
        Calculate contributions to other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]

            remaining = 1.
            fractions = []
            for fraction in self.fractions:
                frac = remaining * fraction.get_value()
                fractions.append(frac)
                remaining -= frac
            fractions.append(remaining)

            if isinstance(self.delay, list):
                for i in range(len(self.to_population)):
                    to_pop = self.to_population[i]
                    scale = fractions[i] * incoming
                    to_pop.update_future_expectation(scale, self.delay[i])
            else:
                for i in range(len(self.to_population)):
                    to_pop = self.to_population[i]
                    scale = fractions[i] * incoming
                    to_pop.update_future_expectation(scale, self.delay)

    def update_data(self):
        """
        Simulate data for other populations by updating their
        future_expectations
        """
        if len(self.from_population.future) > 0:
            incoming = self.from_population.future[0]

            remaining = 1.
            fractions = []
            for fraction in self.fractions:
                frac = remaining * fraction.get_value()
                fractions.append(frac)
                remaining -= frac
            fractions.append(remaining)

            scales = stats.multinomial.rvs(incoming, fractions)
            if isinstance(self.delay, list):
                for i in range(len(self.to_population)):
                    to_pop = self.to_population[i]
                    to_pop.update_future_data(scales[i], self.delay[i])
            else:
                for i in range(len(self.to_population)):
                    to_pop = self.to_population[i]
                    to_pop.update_future_data(scales[i], self.delay)



# -*- coding: utf-8 -*-
"""
Subtractor: A Connector class that 'immediately' removes population.

The purpose of this connector is to include the effect of removal from
populations. For example it is used to track those currently in hospital, instead
of just monitoring the total hospitalized.

@author: karlen
"""

from pypmca.Connector import Connector
from pypmca.Population import Population


class Subtractor(Connector):
    """
    Subtractor removes from the from_population, the newcomers to the
    to_population:
        - connector_name: string, short descriptor
        - from_population: Population object which will be reduced
        by the newcomers to the to_population
        - to_population: Population object having received the newcomers
        of the from_population totals. Its future is not affected by
        this connector.
    """

    def __init__(self, connector_name: str, from_population: Population, to_population: Population):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, Population):
            raise TypeError('Subtractor (' + self.name +
                            ') from_population must be a Population object')

        if not isinstance(to_population, Population):
            raise TypeError('Subtractor (' + self.name +
                            ') to_population must be a Population object')

        if hasattr(from_population, 'report_noise') and from_population.report_noise:
            raise TypeError('Subtractor (' + self.name +
                            ') from_population cannot have report_noise set to True')

        # Identify the from_population as one not to derive daily numbers
        self.from_population.monotonic = False

    def update_expectation(self):
        """
        do subtraction
        """
        self.__do_subtraction()

    def update_data(self):
        """
        do subtraction
        """
        self.__do_subtraction()

    def __do_subtraction(self):
        """
        Subtract new population going into to_population from from_population
        """
        if len(self.to_population.future) > 0:
            reduction = -1 * self.to_population.future[0]
            self.from_population.update_future_fast(reduction)



# -*- coding: utf-8 -*-
"""
Adder: A Connector class that 'immediately' adds population.

The purpose of this connector is to copy incoming population into the
from_population to the to_population.

@author: karlen
"""

from pypmca.Connector import Connector
from pypmca.Population import Population

class Adder(Connector):
    """
    Adder copies the new incoming population into the from_population, 
    to the to_population:
        - connector_name: string, short descriptor
        - from_population: Population object which has newcomers. Its future
        is not affected by this connector.
        - to_population: Population object that receives the newcomers
        entering from_population.
    """

    def __init__(self, connector_name: str, from_population: Population, to_population: Population):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, Population):
            raise TypeError('Adder('+self.name+
                            ') from_population must be a Population object')

        if not isinstance(to_population, Population):
            raise TypeError('Adder ('+self.name+
                            ') to_population must be a Population object')

    def update_expectation(self):
        """
        do addition
        """
        self.__do_addition()

    def update_data(self):
        """
        do subtraction
        """
        self.__do_addition()

    def __do_addition(self):
        """
        Add new population going into from_population into to_population
        """

        newcomers = 0
        if len( self.from_population.future) > 0:
            newcomers = self.from_population.future[0]
        self.to_population.update_future_fast(newcomers)




from sys import version_info

if version_info.major >= 3:
    from pypmca.Population import Population
    from pypmca.Model import Model
    from pypmca.Delay import Delay
    from pypmca.Parameter import Parameter
    from pypmca.Multiplier import Multiplier
    from pypmca.Propagator import Propagator
    from pypmca.Splitter import Splitter
    from pypmca.Adder import Adder
    from pypmca.Subtractor import Subtractor
    from pypmca.Chain import Chain
    from pypmca.Modifier import Modifier
    from pypmca.Injector import Injector
    from pypmca.Ensemble import Ensemble
else:
    raise NotImplementedError('pypmca requires Python 3 or higher. Please upgrade!')

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
"""
Optimizer: Given a model and data, find point estimates for the variable parameters in the model,
along with the covariance (optional)

This is done using the scipy.optimize.curve_fit method

The full_population_name has the prefix daily or total. If daily, then the
differences need to be calculated

The MCMC method relies on calculating a likelihood. The approach was designed around using a cumulative
indicator (reported cases). This was motivated because this indicator suffers from large reporting noise
on a daily basis - with large negative correlations between neighbouring days. The effect is smaller on the
cumulative. Still, reporting noise should be included in simulation samples (it is an optional parameter in pypmca
population models).

Because cumulative data is being used, it is essential to include the auto-covariance in the likelihood.
The shape and normalization of the cumulative indicator are separated so that the auto-covariance
does not include the very large effect from overall normalization. The separation is achieved when
calculating chi^2, by scaling the reference model so that its last point coincides with the simulated/real data.
The contribution from normalization can be evaluated separately by using the values in the last point.
The autocovariance therefore does not include the last point: its rank is one less than the number of points being fit.

It is assumed that the model always has an initial state parameter that scales all final state population sizes.
In the reference models 1&2, this is the cont_0 parameter (number contagious at day 0).
Given that this parameter exists, we can ignore the
contribution to the normalization in evaluating the likelihood.
The simple scaling applied (to separate the shape from normalization) assumes that the initial (t_0)
value for the indicator is near zero.

The following steps are taken to define the likelihood for the MCMC analysis:
(1) The data is fit to find point estimates for the variable parameters. This is done in the analysis tab, and
must be done prior to starting the MCMC analysis. As a check, a naive goodness of fit (shape only) is calculated
and saved as self.chi2d. It is naive, in that it assumes that the data are outcomes of independent Poisson random variables.
The model with the parameters set to these point estimates is referred to as the reference model.

(2) The autocovariance is calculated by producing a set of many simulated datasets from this reference model.
The residuals from the reference model expectations are used to calculate the model auto-covariance. This may differ from the actual
autocovariance if reporting errors are not correctly included or if the model does not contain other important
sources of variance.

(3) An independent set of many simulated datasets from the reference model are produced and the goodness
of fit statistic for each simulated dataset is calculated using different methods to understand the importance of
autocorrelation and to detect any significant difference between noise in data and simulation:
  * chi2_m is the naive goodness of fit (shape only) calculated with respect to the reference model
  * chi2_n is the goodness of fit (normalization only): one degree of freedom - the last data value
  * chi2_f is the naive goodness of fit (shape only) calculated with respect to a reference model fitted to that data
  (this is time consuming and only calculated if self.calc_chi2_f is set to True)
  * chi2_s is the goodness of fit (shape only, with autocorrelation) calculated with respect to the reference model
The means and standard deviations of the gof distributions are saved (in self.chi2m, self.chi2m_sd etc) and the lists
of the values are also available for further study in self.chi2_lists dictionary with keys 'm', 'f', 's'

The chi2_n is calculated with only one degree of freedom (the last data point). This is far from being
well described by a Poisson random variable. Due to fluctuations early in an exponential growth period, the
fluctuations can be very large. Empirically it was found that the distribution for chi2_n/mean(chi2_n) does
follow a chi2 distribution for one degree of freedom. So the likelihood calculation includes, in addition to
the shape contribution, the normalization contribution as calculated by chi2_n/mean(chi2_n), adding one degree
of freedom.

Note: in checking this method with simulated samples, it was found that the chi2_s as calculated by
-2 log(delta likelihood), where likelihood is from the stats.multivariate_normal, follows the chi^2 distribution
much better when the definition is altered (empirically) such that chi2_s = - log(delta likelihood).
Also, the mean of the distribution is much closer to the expected number of degrees of freedom.
It is not understood why this is the case. Since the sampling distribution is better described
and it yields more conservative confidence intervals, the altered statistic is used.

(4) The likelihood is calculated using the probability density of the chi^2 distribution.
The number of degress of freedom can be specified. If the model included all sources of variation correctly,
the best estimate for dof would be the mean(chi2s). This value should be
increased to account for additional degrees of freedom that reflect the unaccounted for noise, coming from the study in part (3).

@author: karlen
"""
import copy
from scipy.optimize import curve_fit
from scipy import stats
import numpy as np


class Optimizer:
    """ Optimizer: point estimator and covariance
    """

    def __init__(self, model, full_population_name, data, data_range):
        self.model = model
        self.full_population_name = full_population_name
        self.population_name = full_population_name[6:]
        self.population_type = full_population_name[:5]
        self.data = data
        self.data_range = data_range
        self.variable_names = []  # float ype variable parameters
        self.variable_initial_values = None
        self.i_variable_names = []  # float ype variable parameters
        self.i_variable_initial_values = None
        self.chi2d = 0.
        self.chi2m = 0.
        self.chi2n = 0.
        self.chi2s = 0.
        self.chi2f = 0.
        self.chi2m_sd = 0.
        self.chi2n_sd = 0.
        self.chi2f_sd = 0.
        self.chi2s_sd = 0.
        self.accept_fraction = 0.
        self.auto_cov = None
        self.chi2_lists = None
        self.calc_chi2f = False

    def func_setup(self):
        # find the variable parameters and return the bounds
        # note: scipy.curve_fit only works for floats
        self.variable_names = []
        self.variable_initial_values = {}
        par_0 = []
        bound_low = []
        bound_high = []
        for par_name in self.model.parameters:
            par = self.model.parameters[par_name]
            if par.get_status() == 'variable':
                if par.parameter_type == 'float':
                    self.variable_names.append(par_name)
                    self.variable_initial_values[par_name] = par.get_value()
                    par_0.append(par.get_value())
                    bound_low.append(par.get_min())
                    bound_high.append(par.get_max())
        return par_0, (bound_low, bound_high)

    def i_func_setup(self):
        self.i_variable_names = []
        self.i_variable_initial_values = {}
        par_0 = []
        bound_low = []
        bound_high = []
        for par_name in self.model.parameters:
            par = self.model.parameters[par_name]
            if par.get_status() == 'variable':
                if par.parameter_type == 'int':
                    self.i_variable_names.append(par_name)
                    self.i_variable_initial_values[par_name] = par.get_value()
                    par_0.append(par.get_value())
                    bound_low.append(par.get_min())
                    bound_high.append(par.get_max())
        return par_0, (bound_low, bound_high)

    def reset_variables(self):
        for var_name in self.variable_names:
            self.model.parameters[var_name].set_value(self.variable_initial_values[var_name])

    def reset_i_variables(self):
        for var_name in self.i_variable_names:
            self.model.parameters[var_name].set_value(self.i_variable_initial_values[var_name])

    def delta(self, cumul):
        diff = []
        for i in range(1, len(cumul)):
            diff.append(cumul[i] - cumul[i - 1])
        return diff

    def i_fit(self):
        """ scan the first integer variable parameter over the range provided
            iterating over different integers by hand
            Returns a dict if a scan was done over an integer,
            describing the scan
        """

        i_par0, i_bounds = self.i_func_setup()
        if len(i_par0) == 0:
            return None

        par_name = self.i_variable_names[0]
        par = self.model.parameters[par_name]
        scan_dict = {'name': par_name}
        val_list = []
        chi2_list = []

        for i in range(par.get_min(), par.get_max() + 1):
            par.set_value(i)
            self.fit()
            val_list.append(i)
            chi2_list.append(self.chi2d)
            self.reset_variables()

        arg = np.argmin(chi2_list)
        par.set_value(val_list[arg])

        scan_dict['val_list'] = val_list
        scan_dict['chi2_list'] = chi2_list
        return scan_dict

    def fit(self, with_cov=False):
        """ work out point estimate. Note that this does not seem to work
            well if the auto_covariance is used - perhaps bug in curve_fit
        """

        def func(x, *params):
            i = -1
            for par_val in params:
                i += 1
                par_name = self.variable_names[i]
                self.model.parameters[par_name].set_value(par_val)

            self.model.reset()
            self.model.evolve_expectations(self.data_range[1])
            pop = self.model.populations[self.population_name]
            func_values = []
            if self.population_type == 'total':
                for xi in x:
                    i = int(round(xi))
                    func_values.append(pop.history[i])
            else:
                diff = self.delta(pop.history)
                for xi in x:
                    i = int(round(xi))
                    func_values.append(diff[i])
            return np.array(func_values)

        par_0, bounds = self.func_setup()
        xdata = np.arange(self.data_range[0], self.data_range[1], 1)

        popt = None
        pcov = None
        if not with_cov:
            popt, pcov = curve_fit(func, xdata, self.data[self.data_range[0]:self.data_range[1]],
                                   p0=par_0, bounds=bounds)
        else:
            popt, pcov = curve_fit(func, xdata, self.data[self.data_range[0]:self.data_range[1]],
                                   p0=par_0, bounds=bounds, sigma=self.auto_cov)

        i = -1
        for par_val in popt:
            i += 1
            par_name = self.variable_names[i]
            self.model.parameters[par_name].set_value(par_val)

        self.model.reset()
        self.model.evolve_expectations(self.data_range[1])
        # report the naive chi^2 (no autocovariance) for shape only.
        # the model is scaled so that it matches the data on the final point
        # see description at top of this file
        scale = self.data[xdata[-1]] / self.model.populations[self.population_name].history[xdata[-1]]
        self.chi2d = 0.
        for x in xdata[:-1]:
            resid = self.data[x] - self.model.populations[self.population_name].history[x] * scale
            self.chi2d += resid ** 2 / (self.model.populations[self.population_name].history[x] * scale)

        return popt, pcov

    def calc_auto_covariance(self, n_rep=100):
        """ Calculate the autocovariance matrix. Do that by generating simulated datasets.

            The autocovariance matrix has rank one less than data since last point is used for normalization scale
            STILL TO DO: deal with daily data. Approach is tuned to study cumulative data for now.***
        """
        # Make a copy first

        sim_model = copy.deepcopy(self.model)
        sim_population = sim_model.populations[self.population_name]
        ref_population = self.model.populations[self.population_name]

        xdata = np.arange(self.data_range[0], self.data_range[1], 1)
        # last point not included: since its residual is zero by definition:
        n_p = len(xdata) - 1
        cov = [[0. for i in range(n_p)] for j in range(n_p)]
        for irep in range(n_rep):
            # produce simulated data -> sim_population histories
            sim_model.reset()
            sim_model.generate_data(self.data_range[1])

            # calculate residuals (data-expectation) for cov
            resid = []
            scale = sim_population.history[xdata[-1]] / ref_population.history[xdata[-1]]

            # last point not used for autocovariance xdata[:-1] removes that point
            for x in xdata[:-1]:
                delta = sim_population.history[x] - ref_population.history[x] * scale
                resid.append(delta)

            for i in range(n_p):
                for j in range(n_p):
                    cov[i][j] += resid[i] * resid[j] / n_rep

        self.auto_cov = cov

    def calc_sim_gof(self, n_rep=100):
        """ Calculate goodness of fit statistics for simulated data:
            chi2m: naive unfitted
            chi2f: naive fitted
            chi2s: unfitted (with autocovariance)
        """

        sim_model = copy.deepcopy(self.model)
        sim_population = sim_model.populations[self.population_name]
        ref_population = self.model.populations[self.population_name]

        # Make a copy for fitting simulated data
        sim_fit_model = copy.deepcopy(self.model)

        chi2_s_list = []
        chi2_m_list = []
        chi2_f_list = []
        chi2_n_list = []

        xdata = np.arange(self.data_range[0], self.data_range[1], 1)
        # last point not included: since its residual is zero by definition:
        n_p = len(xdata) - 1
        zeros = [0. for i in range(n_p)]
        lpdf_zero = stats.multivariate_normal.logpdf(zeros, cov=self.auto_cov)
        for i in range(n_rep):
            sim_model.reset()
            sim_model.generate_data(self.data_range[1])

            scale = sim_population.history[xdata[-1]] / ref_population.history[xdata[-1]]
            resid = []
            chi2_m = 0.
            for x in xdata[:-1]:
                delta = sim_population.history[x] - ref_population.history[x] * scale
                resid.append(delta)
                chi2_m += delta ** 2 / (ref_population.history[x] * scale)

            chi2_m_list.append(chi2_m)
            delta = sim_population.history[xdata[-1]] - ref_population.history[xdata[-1]]
            chi2_n = delta ** 2 / (ref_population.history[xdata[-1]])
            chi2_n_list.append(chi2_n)

            lpdf = stats.multivariate_normal.logpdf(resid, cov=self.auto_cov)
            # empirically it is found that the statistic follows a chi^2 distribution if the formula
            # is altered by changing the nominal 2 in this formula to 1:
            chi2_s_list.append(1. * (lpdf_zero - lpdf))

            if self.calc_chi2f:
                # do a fit to this dataset to see the naive gof for a fitted simulation sample
                # this needs to be turned on by the inquisitive user
                # ----------------------------------------------------------------------------
                sim_optimizer = Optimizer(sim_fit_model, self.full_population_name, sim_population.history,
                                          self.data_range)
                sim_optimizer.reset_variables()
                sim_popt, sim_pcov = sim_optimizer.fit()
                sim_ref_population = sim_fit_model.populations[self.population_name]

                # calculate residuals (data-expectation) for cov and chi^2
                resid_f = []
                scale_f = sim_population.history[xdata[-1]] / sim_ref_population.history[xdata[-1]]
                chi2_f = 0.
                # last point not used for autocovariance xdata[:-1] removes that point
                for x in xdata[:-1]:
                    delta = sim_population.history[x] - sim_ref_population.history[x] * scale_f
                    resid_f.append(delta)
                    chi2_f += delta ** 2 / (sim_ref_population.history[x] * scale_f)

                chi2_f_list.append(chi2_f)

        self.chi2m = np.mean(chi2_m_list)
        self.chi2m_sd = np.std(chi2_m_list)
        self.chi2n = np.mean(chi2_n_list)
        self.chi2n_sd = np.std(chi2_n_list)
        if self.calc_chi2f:
            self.chi2f = np.mean(chi2_f_list)
            self.chi2f_sd = np.std(chi2_f_list)
        self.chi2s = np.mean(chi2_s_list)
        self.chi2s_sd = np.std(chi2_s_list)
        # for further study of gof distributions:
        self.chi2_lists = {'m': chi2_m_list, 'f': chi2_f_list, 's': chi2_s_list, 'n': chi2_n_list}

    def mcmc(self, n_dof, chi2n_mean, n_MCMC):
        """ Make a MCMC chain (n_MCMC points) assuming n_dof defines gof statistic for data.
            The gof is calculated by shape (many dof) and normalization (one dof). The latter calculation
            requires a scaling factor: the mean chi2_n.
            Currently only setup to deal with float variables, but a simple extension would allow int/bool
            Method returns chain in array of dictionaries
        """

        var_names = []
        var_parameters = {}
        var_values = {}
        prior_function = {}
        hypercube = {}
        mean = {}
        half_width = {}
        sigma = {}

        sim_model = copy.deepcopy(self.model)
        sim_population = sim_model.populations[self.population_name]

        for par_name in sim_model.parameters:
            par = sim_model.parameters[par_name]
            if par.get_status() == 'variable':
                var_names.append(par_name)
                var_parameters[par_name] = par
                var_values[par_name] = par.get_value()
                prior_function[par_name] = par.prior_function
                hypercube[par_name] = par.mcmc_step
                mean[par_name] = par.prior_parameters['mean']
                if par.prior_function == 'uniform':
                    half_width[par_name] = par.prior_parameters['half_width']
                else:
                    sigma[par_name] = par.prior_parameters['sigma']

        xdata = np.arange(self.data_range[0], self.data_range[1], 1)
        zeros = [0. for i in range(len(xdata)-1)]
        lpdf_zero = stats.multivariate_normal.logpdf(zeros, cov=self.auto_cov)

        # calculate log of posterior probability parameters
        def logP():
            # check if selected parameters within bounds (uniform priors)
            for var_name in var_names:
                if prior_function[var_name] == 'uniform':
                    par = var_parameters[var_name]
                    if np.abs(par.get_value() - mean[var_name]) > half_width[var_name]:
                        return -np.inf
                else:
                    par_val = par.get_value()
                    if par_val < par.get_min() or par_val > par.get_max():
                        return -np.inf

            sim_model.reset()
            sim_model.evolve_expectations(self.data_range[1])

            resid = []
            # do not use the final point
            for x in xdata[:-1]:
                resid.append(self.data[x] - sim_population.history[x])

            lpdf = stats.multivariate_normal.logpdf(resid, cov=self.auto_cov)
            # empirical finding that the statistic follows chi^2 if the
            # nominal 2 in the following form is replaced by 1.
            chi2 = 1. * (lpdf_zero - lpdf)

            # add to the shape chi^2, the normalization chi^2 scaled by chi2n_mean
            # this corresponds to one extra degree of freedom
            delta = self.data[xdata[-1]] - sim_population.history[xdata[-1]]
            chi2_n = delta ** 2 / (sim_population.history[xdata[-1]])
            chi2 += chi2_n/chi2n_mean

            ln_P = stats.chi2.logpdf(chi2, n_dof)

            # Deal with any normal priors
            for var_name in var_names:
                if prior_function[var_name] == 'normal':
                    par = var_parameters[var_name]
                    dev = (par.get_value() - mean[var_name]) / sigma[var_name]
                    ln_P -= dev * dev / 2.

            return ln_P

        chain = []

        n_accept = 0
        lp_curr = logP()
        for i in range(n_MCMC):
            new_var_values = var_values.copy()
            for var_name in var_names:
                new_var_values[var_name] += hypercube[var_name] * (1. - 2. * stats.uniform.rvs())
                par = var_parameters[var_name]
                par.set_value(new_var_values[var_name])

            lp_new = logP()
            if lp_new != -np.inf:
                del_lp = lp_new - lp_curr
                if del_lp > -30:
                    if del_lp > 0 or (stats.uniform.rvs() < np.exp(del_lp)):
                        var_values = new_var_values
                        lp_curr = lp_new
                        n_accept += 1
            chain.append(var_values)

        self.accept_fraction = 1. * n_accept / n_MCMC
        return chain



# -*- coding: utf-8 -*-
"""
Model: A class consisting of a list of populations and connectors.

With this information, the evolution of each population is performed,
either by updating expectations or simulated data.

@author: karlen
"""
from datetime import date
import pickle
from pathlib import Path
from pypmca.Connector import Connector
from pypmca.Population import Population
from pypmca.Transition import Transition
from pypmca.Parameter import Parameter

class Model:
    """
    Class that defines populations and their evolution through connectors
    """

    def __init__(self, model_name):

        self.name = str(model_name)
        self.description = ''
        self.t0 = date(2020, 3, 1)
        # default time step is one day. (hidden)
        self.__time_step = 1.
        self.populations = {}
        self.connectors = {}
        self.connector_list = []
        self.transitions = {}
        self.parameters = {}
        self.delays = {}
        self.lnl = None
        self.population_mcmc_time_step_range = {}

        self.boot_pars = {}
        self.boot_needed = True

    def set_t0(self, year=2020, month=1, day=1):
        """Define the date of the first element in the history lists
        """
        self.t0 = date(year, month, day)

    def set_time_step(self, time_step):
        """
        Define the length of time corresponding to one step in the evolution.
        Update all delays and transitions if this is changed.

        Parameters
        ----------
        time_step : size of time step in units of days. So a 1 hour time step
                    is specified as 1./24.

        Returns
        -------
        None.

        """
        if not isinstance(time_step, float):
            raise TypeError('Model (' + self.name +
                            ' time_step must be a float')
        if not time_step > 0.:
            raise ValueError('Model (' + self.name +
                             ' time_step must be positive')
        self.__time_step = time_step

        # update the objects that are sensitive to time_step
        for delay_name in self.delays:
            delay = self.delays[delay_name]
            delay.update()
        for transition_name in self.transitions:
            transition = self.transitions[transition_name]
            transition.update()
        for connector_name in self.connectors:
            connector = self.connectors[connector_name]
            if type(connector).__name__ == 'Chain':
                connector.update()

    def get_time_step(self):
        """
        Get the length of time corresponding to one step in the evolution

        Returns
        -------
        real value for one time step

        """
        return self.__time_step

    def get_history(self, population_name):
        if not isinstance(population_name, str):
            raise TypeError('model.get_history argument must be a str')

        pop = self.populations[population_name]
        return pop.history

    def boot_setup(self, boot_population, boot_value, exclusion_populations=None):
        """
        Typically most populations are initialized to zero. Propogation and
        growith requires at least one population to not be zero. However,
        starting a model with large values in a source population can lead to
        unusual behaviour for initial time steps, as that would be inconsistent
        with the zeros in most other populations. On the other hand, starting
        with small values in a source population can lead to extreme sensitivity to
        time. To solve this issue, we boot the model, by starting with
        small populations and evolve the system until a population exceeds
        a target value. History is removed and the current populations and futures
        are used for the initial state of the model. To allow for
        continuous dependence, the current value and futures are scaled by the ratio of
        the target value to actual value, except for the exclusion_populations.
            - boot_population: str, source population
            - boot_value: float or int, small value to begin boot sequence.
            - exclusion_populations: str or str-list of populations to exclude
            from current value scaling. The history is set to the initial value for
            the exclusion populations, the future contributions are scaled.
        The boot sequence ends when the boot_population reaches or exceeds
        the value set for the initial_value of that population.

        To fit data to a model it will be necessary to set the initial size of
        a critical population (the boot population) to match the scale for data.
        For that reason, we insist that the boot population be initialized by
        a parameter below.

        One downside of the rescaling is that the long term of contagious
        population can be non-zero (a few).

        It can be useful to save a copy of a booted model, and perform restarts
        of that model.
        """

        name_boot_pop = str(boot_population)
        if name_boot_pop not in self.populations:
            raise ValueError('Error in boot_setup for model (' + self.name +
                             '). ' + name_boot_pop + 'is not present in model.')
        initial_value = self.populations[name_boot_pop].initial_value
        if not isinstance(initial_value, Parameter):
            raise ValueError('Error in boot_setup for model (' + self.name +
                             '). ' + name_boot_pop + 'is not initialized by a parameter.')
        exc_pop_list = []
        if exclusion_populations is not None:
            if isinstance(exclusion_populations, list):
                for exc_pop in exclusion_populations:
                    name_exc_pop = str(exc_pop)
                    if name_exc_pop not in self.populations:
                        raise ValueError('Error in boot_setup for model (' + self.name +
                                         '). ' + name_exc_pop + 'is not present in model.')
                    exc_pop_list.append(name_exc_pop)
            else:
                name_exc_pop = str(exclusion_populations)
                if name_exc_pop not in self.populations:
                    raise ValueError('Error in boot_setup for model (' + self.name +
                                     '). ' + name_exc_pop + 'is not present in model.')
                exc_pop_list.append(name_exc_pop)

        self.boot_pars['boot_population'] = name_boot_pop
        self.boot_pars['boot_value'] = boot_value
        self.boot_pars['exclusion_populations'] = exc_pop_list

    def reset(self):
        """
        Reinitialize all populations, eliminating history and future, and
        setting initial values. Also reset transitioned parameters
        """
        for key in self.populations:
            self.populations[key].reset()

        # do this in reverse order, so earlier transitions get precidence
        trans_dict = {}
        rc = 0.
        for key in self.transitions:
            rc += 0.01
            trans = self.transitions[key]
            trig_step = trans.trigger_step + rc
            trans_dict[trig_step] = trans
        trans_list = list(trans_dict.keys())
        trans_list.sort()
        trans_list.reverse()
        for trans_step in trans_list:
            trans = trans_dict[trans_step]
            trans.reset()

        self.boot_needed = True

    def boot(self, expectations=True):
        """ Using information in boot_pars, evolve the system to a point where
        the target population is reached or exceeded:
            - expectations: if True, after the boot completes, the
            history will begin with an expectation,
            otherwise, a Poisson number is drawn using that expectation.
        Note: check is added so that if boot requirement is not made within
        10k steps, it fails (rather than stay in infinite loop)
        """
        if len(self.boot_pars) == 0:
            raise RuntimeError('Model boot parameters not set.')

        # start from a clean slate of populations set to their initial values
        self.reset()

        # need to set this to False before evolving expectations
        self.boot_needed = False

        boot_pop = self.populations[self.boot_pars['boot_population']]
        goal_value = boot_pop.history[-1]
        boot_pop.history[-1] = self.boot_pars['boot_value']

        # exclusion_populations are reset to their original values after the
        # boot. Save them here.
        exclusions = self.boot_pars['exclusion_populations']
        exc_vals = {}
        for exc_pop_name in exclusions:
            exc_pop = self.populations[exc_pop_name]
            exc_vals[exc_pop_name] = exc_pop.history[-1]

        last_value = -1
        i_step = 0
        while boot_pop.history[-1] < goal_value and boot_pop.history[-1] >= last_value:
            last_value = boot_pop.history[-1]
            self.evolve_expectations(1)
            i_step += 1
            if i_step % 1000 == 0:
                if boot_pop.history[-1] < 0.1 * goal_value:
                    self.boot_needed = True
                    raise ValueError('The boot process is taking too long to complete. ' +
                                     'Adjust parameters to allow the goal to be reached.')

        if boot_pop.history[-1] < goal_value:
            raise ValueError('The boot process did not reach its target. Consider reduce the target.')

        scale = goal_value / boot_pop.history[-1]

        # Erase history until current. Scale all histories and futures.
        for key in self.populations:
            pop = self.populations[key]
            pop.remove_history()
            if exclusions is None or str(pop) not in exclusions:
                pop.scale_history(scale, expectations)
            elif exclusions is not None and str(pop) in exclusions:
                pop.history[-1] = exc_vals[str(pop)]
            pop.scale_future(scale, expectations)

        # restore goal_value for data production
        if not expectations:
            boot_pop.history[-1] = int(goal_value)

    def evolve_expectations(self, n_step):
        """
        First check if a boot is needed: this is required if
        reset_populations is called

        Take n_steps of
        0) check if it is time to apply a transition
        1) updating futures
        2) appending new value to history array
        3) remove current time step in future array
        """
        if self.boot_needed:
            self.boot(expectations=True)

        for step in range(n_step):
            self.do_transitions(step, expectations=True)

            # calculate future expectations
            self.calculate_future(expectations=True)

            # make one time step
            self.do_time_step(expectations=True)

    def generate_data(self, n_step):
        """
        Produce data
        """
        if self.boot_needed:
            self.boot(expectations=False)

        for step in range(n_step):
            self.do_transitions(step, expectations=False)

            # calculate future data
            self.calculate_future(expectations=False)

            # make one time step
            self.do_time_step(expectations=False)

    def do_transitions(self, step, expectations=True):
        for transition_name in self.transitions:
            transition = self.transitions[transition_name]
            if step == transition.trigger_step:
                if transition.enabled:
                    transition.take_action(expectations)

    def calculate_future(self, expectations=True):
        for connector_name in self.connector_list:
            connector = self.connectors[connector_name]
            if expectations:
                connector.update_expectation()
            else:
                connector.update_data()

    def do_time_step(self, expectations=True):
        for key in self.populations:
            self.populations[key].do_time_step(expectations)

    def add_transition(self, transition):
        """
        Add a transition to the model. Special action is taken in the specified
        time step. Transitions do not apply to the boot phase.
        """
        if not isinstance(transition, Transition):
            raise TypeError('Error in adding transition to ' + self.name +
                            ': add_transition argument must be a Transition object')
        if str(transition) in self.transitions:
            raise ValueError('Error in adding transition (' +
                             str(transition) + ') to model (' + self.name +
                             '). Transition with that name already present in model.')
        self.transitions[str(transition)] = transition

        self.__update_parameter_list(transition)

    def add_connector(self, connector, before_connector=None, after_connector=None):
        """
        Add a connector to the model. The model is evolved by processing
        the connectors in the order that they appear in connector_list.
        Connectors must have unique names.

        Parameters
        ----------
        connector : Connector object
        """
        if not isinstance(connector, Connector):
            raise TypeError('Error in adding connector to ' + self.name +
                            ': add_connector argument must be a Connector object')
        if str(connector) in self.connectors:
            raise ValueError('Error in adding connector (' +
                             str(connector) + ') to model (' + self.name +
                             '). Connector with that name already present in model.')
        if after_connector is not None and before_connector is not None:
            raise ValueError('Error in adding connector(' +
                             str(connector) + ') to model (' + self.name +
                             '). Cannot specify both before and after locations.')
        index_loc = -1
        if before_connector is not None:
            name = str(before_connector)
            if name not in self.connector_list:
                raise ValueError('Error in adding connector(' +
                                 str(connector) + ') to model (' + self.name +
                                 '): ' + str(before_connector) + ' not found.')
            index_loc = self.connector_list.index(name)
        elif after_connector is not None:
            name = str(after_connector)
            if name not in self.connector_list:
                raise ValueError('Error in adding connector(' +
                                 str(connector) + ') to model (' + self.name +
                                 '): ' + str(after_connector) + ' not found.')
            index_loc = self.connector_list.index(name) + 1
        else:
            index_loc = len(self.connector_list)

        self.connector_list.insert(index_loc, str(connector))
        self.connectors[str(connector)] = connector

        self.__update_population_list(connector)
        self.__update_parameter_list(connector)
        self.__update_delay_list(connector)

    def remove_connector(self, connector):
        """
        Remove a connector from the model.

        Parameters
        ----------
        connector : Connector or str
            Connector to be removed. Can be specified by its name.

        Returns
        -------
        removed_connector : Connector
            Returned so that it can be easily added back.
        """

        name = str(connector)
        if name not in self.connector_list:
            raise ValueError('Error in removing connector(' +
                             str(connector) + ') from model (' + self.name +
                             '): ' + str(connector) + ' not found.')

        self.connector_list.remove(name)
        removed_connector = self.connectors.pop(name, None)

        self.update_lists()
        return removed_connector

    def remove_transition(self, transition):
        name = str(transition)
        if name not in self.transitions:
            raise ValueError('Error in removing transition(' +
                             str(transition) + ') from model (' + self.name +
                             '): ' + str(transition) + ' not found.')

        removed_transition = self.transitions.pop(name, None)

        self.update_lists()
        return removed_transition

    def update_lists(self):
        """ After modifying any element of the model that involves
            populations or parameters, call this to be sure that
            the list of active populations and parameters is current
        """
        # remake list of active populations and their parameters
        self.populations = {}
        self.parameters = {}
        self.delays = {}

        for con_name in self.connector_list:
            con = self.connectors[con_name]
            self.__update_population_list(con)
            self.__update_parameter_list(con)
            self.__update_delay_list(con)

        # add to list of active parameters
        for trans_name in self.transitions:
            trans = self.transitions[trans_name]
            self.__update_parameter_list(trans)

    def __update_parameter_list(self, obj):

        if type(obj).__name__ == 'Injector':
            self.__add_to_parameter_list(obj.transition_time)
            self.__add_to_parameter_list(obj.injection)

        elif type(obj).__name__ == 'Modifier':
            self.__add_to_parameter_list(obj.transition_time)
            self.__add_to_parameter_list(obj.parameter_before)
            self.__add_to_parameter_list(obj.parameter_after)

        elif isinstance(obj, Connector):
            con = obj
            for par_key in con.parameters:
                par = con.parameters[par_key]
                self.__add_to_parameter_list(par)
            if type(con).__name__ == 'Chain':
                for link_con in con.chain:
                    for link_par_key in link_con.parameters:
                        link_par = link_con.parameters[link_par_key]
                        self.__add_to_parameter_list(link_par)

        elif isinstance(obj, Population):
            pop = obj
            for par_key in pop.parameters:
                par = pop.parameters[par_key]
                self.__add_to_parameter_list(par)

    def __add_to_parameter_list(self, obj):
        # check that the name of the parameter is unique
        if str(obj) in self.parameters:
            par = self.parameters[str(obj)]
            if obj != par:
                raise ValueError('Two parameters share the same name: ' + str(obj))
        else:
            self.parameters[str(obj)] = obj

    def __update_population_list(self, connector):
        # make a list of populations involved in this connector
        pops = []
        args = [connector.from_population, connector.to_population]
        for arg in args:
            if isinstance(arg, list):
                for pop in arg:
                    pops.append(pop)
            else:
                pops.append(arg)
        if hasattr(connector, 'chain'):
            chain = connector.chain
            for link in chain:
                pops.append(link.from_population)
                pops.append(link.to_population)

        # If necessary add to population dictionary.
        # Check there are no duplicate keys
        for pop in pops:
            key = pop.name
            if key in self.populations:
                dict_pop = self.populations[key]
                if dict_pop != pop:
                    raise ValueError('Two populations share the same name: ' +
                                     pop.name)
            else:
                self.populations[key] = pop
                self.__update_parameter_list(pop)

    def __update_delay_list(self, connector):
        # make a list of delays involved in this connector
        delays = []
        if hasattr(connector, 'delay'):
            if isinstance(connector.delay, list):
                for delay in connector.delay:
                    delays.append(delay)
                    if type(connector).__name__ == 'Chain':
                        for chain_con in connector.chain:
                            delays.append(chain_con.delay)

            else:
                delays.append(connector.delay)
                if type(connector).__name__ == 'Chain':
                    for chain_con in connector.chain:
                        delays.append(chain_con.delay)

        # If necessary add to delay dictionary.
        # Check there are no duplicate keys
        for delay in delays:
            key = delay.name
            if key in self.delays:
                dict_delay = self.delays[key]
                if dict_delay != delay:
                    raise ValueError('Two delays share the same name: ' +
                                     delay.name)
            else:
                self.delays[key] = delay

    def save_file(self, filename):
        """
        Save a copy of the current model to a file. The model can be restored
        at a late time using Model.open_file(filename).

        Before saving, the model futures are removed, and the initial_values
        are set to the current values for all parameters.

        If no extension is provided, the default extension, .pypm is added.

        Parameters
        ----------
        filename : Path or str
            name of file to save model

        Returns
        -------
        None.

        """

        try:
            filepath = Path(filename).resolve()
        except:
            raise TypeError('Input arg could not be converted to a valid path: {}' +
                            '\n It must be a str or Path-like.'.format(filename))

        if len(filepath.suffix) < 2:
            filepath = filepath.with_suffix('.pypm')

        #       eliminate histories, undo transitioned parameter adjustments

        self.reset()

        #       set initial parameter values to their current values to define a new
        #       reference

        for par_name in self.parameters:
            par = self.parameters[par_name]
            par.new_initial_value()

        with open(filepath, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)


    @classmethod
    def open_file(cls, filepath):
        """
        Restore a model that was save to a file using Model.save_file(filename)

        Parameters
        ----------
        filepath : Path or str
            name of existing model file to open

        Returns
        -------
        Model
            The model object saved in the file

        """

        try:
            filepath = Path(filepath).resolve()
        except:
            raise TypeError('Input arg could not be converted to a valid path: {}' +
                            '\n It must be a str or Path-like.'.format(filepath))

        if not filepath.exists():
            raise ValueError('Filepath does not exist: {}'.format(filepath))

        if len(filepath.suffix) < 2:
            filepath = filepath.with_suffix('.pypm')

        with open(filepath, 'rb') as f:
            return pickle.load(f)



# -*- coding: utf-8 -*-
"""
Parameter: A class for defining parameters of the model:

Parameters with parameter_status = 'variable' identifies parameters that are to
be adjusted within optimization or MCMC exploration methods. For such parameters,
priors must be supplied using parameter_function and prior_parameters.

Parameters with parameter_status = 'fixed' can be adjusted manually in an
interactive exploration of model parameters. 

The automatic transition of a parameter at defined point of the evolution is
done by adding a Modifier object to the Model.

The pre-calculated delay distributions need to be recalculated when its
parameters are changed. Such parameters have must_update = True

@author: karlen
"""


class Parameter:
    """
    Model parameter:
    - parameter_name: short descriptor
    - description: long form description for documentation
    - value: value to be set initially or when reset requested
    - parameter_min, parameter_max, allowed range for int or float parameter
    - parameter_type: 'int', 'float', 'bool'
    - parameter_status: 'fixed', 'variable'
    - prior_function: if (real or integer) and variable, the functional form
      of prior. Allowed: 'norm' or 'uniform'
    - prior_parameters for (list of floats, not parameter objects):
        - for 'norm': dict [mean, sigma]
        - for 'uniform': dict [mean, half_width]
        - for boolean: truth probability
    - mcmc_step is maximum size of step during MCMC ((1-2*uni_ran)*step)
    """

    PARAMETER_TYPES = {'int': int, 'float': float, 'bool': bool}
    PARAMETER_STATUSES = ['fixed', 'variable']
    PRIOR_FUNCTIONS = ['norm', 'uniform']
    PRIOR_PARS = {'norm': ['mean', 'sigma'],
                  'uniform': ['mean', 'half_width']}

    def __init__(self, parameter_name, value,
                 parameter_min=0, parameter_max=1, description='',
                 parameter_type='float', parameter_status='fixed',
                 prior_function=None, prior_parameters=None, mcmc_step=None, hidden=True):

        if parameter_name.find(',') > -1:
            raise ValueError('Error in constructing ' + self.name +
                             ': name cannot contain a comma.')
        self.name = str(parameter_name)
        self.must_update = False
        self.parents = []

        if parameter_type not in self.PARAMETER_TYPES:
            buff = '/'.join(self.PARAMETER_TYPES)
            raise ValueError('Parameter (' + self.name + ') type "' +
                             str(parameter_type) +
                             '" invalid: not in: ' + buff)
        self.parameter_type = parameter_type

        self.description = description
        self.parameter_min = parameter_min
        self.parameter_max = parameter_max
        self.__value = None
        self.set_value(value)
        self.initial_value = value

        if parameter_status not in self.PARAMETER_STATUSES:
            buff = '/'.join(self.PARAMETER_STATUSES)
            raise ValueError('Parameter (' + self.name + ') status "' +
                             str(parameter_type) +
                             '" invalid: not in: ' + buff)
        self.__status = parameter_status
        self.__status_changed = False

        self.prior_function = None
        self.prior_parameters = None
        if parameter_status == 'variable':
            self.set_variable(prior_function, prior_parameters)

        if not isinstance(hidden, bool):
            raise TypeError('Parameter (' + self.name + ') hidden argument ' +
                            'must be of type bool')
        self.hidden = hidden

        self.mcmc_step = mcmc_step

    def __str__(self):
        return self.name

    def get_min(self):
        return self.parameter_min

    def get_max(self):
        return self.parameter_max

    def set_min(self, val):
        if self.parameter_type == 'int':
            self.parameter_min = int(val)
        else:
            self.parameter_min = val

    def set_max(self, val):
        if self.parameter_type == 'int':
            self.parameter_max = int(val)
        else:
            self.parameter_max = val

    def reset(self):
        self.set_value(self.initial_value)

    def new_initial_value(self):
        self.initial_value = self.get_value()

    def get_value(self):
        """
        Return current value of parameter

        """
        return self.__value

    def set_value(self, new_value):
        """
        Change the value of the parameter

        Parameters
        ----------
        new_value : TYPE must match parameter_type
            new value of the parameter

        Returns
        -------
        None.

        """
        value_type = type(new_value).__name__
        # avoid issues with float64 vs float
        if value_type[:len(self.parameter_type)] != self.parameter_type:
            raise TypeError('Parameter (' + self.name +
                            ') value type (' + type(new_value).__name__ +
                            ') does not match parameter_type (' +
                            self.parameter_type + ')')
        self.__value = new_value

        # if this parameter is used for delay - do a recalculation of the distribution
        if self.must_update:
            for parent in self.parents:
                parent.update()

    def set_must_update(self, parent):
        """ identify this parameter as one that changes to its value requires 
        that a calculation in the parent needs to be updated
        """
        self.must_update = True
        if parent not in self.parents:
            self.parents.append(parent)

    def get_status(self):
        return self.__status

    def set_fixed(self):
        """
        Change the parameter status from 'variable' to 'fixed'.

        Returns
        -------
        None.

        """
        self.__status = 'fixed'
        self.__status_changed = True

    def set_variable(self, prior_function=None, prior_parameters=None):
        """
        Change the parameter status from 'fixed' to 'variable'. If the
        parameter had previously been 'variable', and the prior is not
        specified, the prior previously used is reapplied by default.

        Parameters
        ----------
        prior_function : STR, optional
            DESCRIPTION. if (real or integer) and variable, the functional form
            of prior. Allowed: 'norm' or 'uniform'
        prior_parameters : dict, optional
            DESCRIPTION. (list of floats, not parameter objects):
                - for 'norm': dict  having keys ['mean', 'sigma', 'step']
                - for 'uniform': dict having keys ['mean', 'half_width', 'step']
                - for boolean: float, truth probability

        if prior not supplied, the previously used prior will be used
        if prior not supplied and no prior previously used, then set
        prior to be uniform: [parameter_min, parameter_max]

        Note that the priors are used for MCMC. For the fitting, the
        bounds are set directly by paramete_min, parameter_max

        Returns
        -------
        None.

        """

        # by default use the previously used prior or uniform if no previous prior
        if prior_function is None and prior_parameters is None:
            if self.prior_function is None or self.prior_parameters is None:
                self.prior_function = 'uniform'
                mean = (self.parameter_max + self.parameter_min) / 2
                half_width = (self.parameter_max - self.parameter_min) / 2
                self.prior_parameters = {'mean': mean, 'half_width': half_width}

        else:

            if self.parameter_type != 'bool':
                if prior_function is None:
                    raise TypeError('Variable parameter (' + self.name +
                                    ') prior_function cannot be None')
                if prior_function not in self.PRIOR_FUNCTIONS:
                    buff = '/'.join(self.PRIOR_FUNCTIONS)
                    raise ValueError('Parameter (' + self.name + ') prior_function "' +
                                     prior_function +
                                     '" invalid: not in: ' + buff)
                self.prior_function = prior_function

                if prior_parameters is None:
                    raise TypeError('Variable parameter (' + self.name +
                                    ') prior_parameters cannot be None')
                if not isinstance(prior_parameters, dict):
                    raise TypeError('Variable parameter (' + self.name +
                                    ') prior_parameters must be a dictionary')
                for key in self.PRIOR_PARS[prior_function]:
                    if key not in prior_parameters:
                        raise ValueError('Variable parameter (' + self.name +
                                         ') prior_parameters must include ' + key)
                    if not isinstance(prior_parameters[key], float):
                        raise TypeError('Variable parameter (' + self.name +
                                        ' prior_parameters must be floats')

            elif self.parameter_type == 'bool':
                if self.prior_parameters is None:
                    raise TypeError('Variable parameter (' + self.name +
                                    ') prior_parameters cannot be None')
                if not isinstance(prior_parameters, float):
                    raise TypeError('Variable parameter (' + self.name +
                                    ' prior_parameters must be a float')
            self.prior_parameters = prior_parameters

        self.__status = 'variable'
        self.__status_changed = True



#!/usr/bin/env python

"""Tests for `pypmca` package."""

import pytest

from pypmca import Model, Population, Delay, Parameter, Multiplier, Propagator, \
    Splitter, Adder, Subtractor, Chain, Modifier, Injector
import numpy as np


def test_class_Model():
    """tests to ensure the behaviour class Model"""
    test_model = Model('test')


def test_class_Population():
    """tests to ensure the behaviour class Population"""
    init_value = 100
    test_pop = Population('test population', init_value, description='For testing populations',
                          hidden=True, color='black', show_sim=False, report_noise=False,
                          report_noise_par=None)
    assert test_pop.history[0] == init_value

    incoming = 10
    test_pop.update_future_fast(incoming)
    assert test_pop.future[0] == incoming

    test_pop.do_time_step()
    assert test_pop.history[1] == init_value + incoming
    assert len(test_pop.future) == 0 or test_pop.future == 0

    scale_factor = 0.5
    test_pop.scale_history(scale_factor)
    assert test_pop.history[0] == init_value * scale_factor
    assert test_pop.history[1] == (init_value + incoming) * scale_factor

    # check that noise in reporting makes sense
    # Expectations are not effected - data should be
    # but the total reported should be the same after all reporting complete
    noise_factor = Parameter('noise_factor', 0.3)
    # restart - back to initial value
    for expectations in [True, False]:
        test_pop.reset()
        future = [0, 10, 100, 100, 100, 10, 0, 0, 0]
        future_sum = np.sum(np.array(future))
        test_pop.set_report_noise(True, noise_factor)
        test_pop.future = future
        for i in range(len(future) + 5):
            test_pop.do_time_step(expectations=expectations)
        assert test_pop.history[-1] == init_value + future_sum


def test_class_Parameter():
    """tests to ensure the behaviour class Parameter"""
    test_parameter = Parameter('test_par', 10., parameter_min=5., parameter_max=100.)
    for value in [-10., 500., 50]:
        error_caught = False
        try:
            test_parameter.set_value(value)
        except ValueError:
            error_caught = True
        except TypeError:
            error_caught = True
        assert error_caught


def test_class_Delay():
    """tests to ensure the behaviour class Delay"""
    test_model = Model('test_model')
    EPS = 0.1
    mean = 10.
    std_dev = 4.
    half_width = float(std_dev * np.sqrt(12.) / 2.)
    k_vals = [1, 2, 3]
    for delay_type in ['norm', 'uniform', 'erlang']:
        k_s = [1]
        if delay_type == 'erlang':
            k_s = k_vals
        for k in k_s:
            delay_pars = {
                'mean': Parameter('mean', mean, parameter_min=-100., parameter_max=100.),
                'sigma': Parameter('sigma', std_dev, parameter_min=-100., parameter_max=100.),
                'half_width': Parameter('hw', half_width, parameter_min=-100., parameter_max=100.),
                'k': Parameter('k', k, parameter_type='int', parameter_min=1, parameter_max=100)
            }
            for time_step in [1., 1. / 4.]:
                test_model.set_time_step(time_step)
                # The delay is created after the model: since it is not associated with a connector, the model
                # does not know to call its update method. (The model does not have stand alone delays.)
                test_delay = Delay('test_delay', delay_type, delay_parameters=delay_pars, model=test_model)
                distribution = test_delay.future_expectations
                sum_p = 0.
                sum_tp = 0.
                sum_ttp = 0.
                for i in range(len(distribution)):
                    sum_p += distribution[i]
                    sum_tp += i * distribution[i] * time_step
                    sum_ttp += i * i * distribution[i] * time_step ** 2
                assert np.abs(sum_p - 1.) < EPS
                est_mean = sum_tp
                assert np.abs(est_mean - mean) < EPS
                est_sigma = np.sqrt(sum_ttp - sum_tp * sum_tp)
                if delay_type != 'erlang':
                    assert np.abs(est_sigma - std_dev) < EPS
                else:
                    assert np.abs(est_sigma - mean / np.sqrt(1. * k)) < EPS


def test_class_Adder():
    """tests to ensure the behaviour class Adder"""
    from_init = 20
    to_init = 10
    from_pop = Population('from_pop', from_init)
    from_next = 40
    from_pop.future = [from_next]
    to_pop = Population('to_pop', to_init)
    test_adder = Adder('test_add', from_pop, to_pop)
    test_adder.update_expectation()
    assert to_pop.future[0] == from_next


def test_class_Injector():
    """tests to ensure the behaviour class Injector"""
    test_model = Model('test_model')
    number = 50.
    inject = Parameter('inject', number, parameter_min=0., parameter_max=1000.)
    time = 5
    trans_time = Parameter('time', time, parameter_type='int', parameter_min=0, parameter_max=1000)
    to_pop = Population('to_pop', 0)
    test_injector = Injector('injector', 'rel_days', trans_time, to_pop, inject, model=test_model)
    test_model.add_transition(test_injector)
    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        to_pop.reset()
        test_injector.take_action()
        assert to_pop.future[0] == number
        assert np.abs(test_injector.trigger_step - time / time_step) < 0.1


def test_class_Modifier():
    """tests to ensure the behaviour class Modifier"""
    test_model = Model('test_model')
    mod_time = Parameter('time', 5, parameter_type='int', parameter_min=0, parameter_max=1000)
    par_val = 0.3
    par_0_val = 0.5
    par_1_val = 0.7
    parameter = Parameter('par', par_val)
    parameter_0 = Parameter('par_0', par_0_val)
    parameter_1 = Parameter('par_1', par_1_val)
    test_modifier = Modifier('test_modifier', 'rel_days', mod_time, parameter, parameter_0, parameter_1,
                             model=test_model)
    test_model.add_transition(test_modifier)
    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        parameter.reset()
        assert parameter.get_value() == par_val
        test_modifier.take_action()
        assert parameter.get_value() == par_1_val
        test_modifier.reset()
        assert parameter.get_value() == par_0_val
        assert np.abs(test_modifier.trigger_step - 5 / time_step) < 0.1


def test_class_Multiplier():
    """tests to ensure the behaviour class Multiplier"""
    test_model = Model('test_model')
    EPS = 1.
    n1 = 50.
    n2 = 20.
    n3 = 2.
    scale = 0.1
    f_pops = [Population('f1_pop', n1), Population('f2_pop', n2), Population('f3_pop', n3)]
    to_pop = Population('to_pop', 0.)
    scale_par = Parameter('alpha', scale)
    delay = Delay('fast', 'fast')
    test_multiplier = Multiplier('test_multiplier', f_pops, to_pop, scale_par, delay, model=test_model)
    test_model.add_connector(test_multiplier)
    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        # expectation:
        expected = n1 * n2 / n3 * scale * time_step
        to_pop.reset()
        test_multiplier.set_distribution('poisson', None)
        test_multiplier.update_expectation()
        assert to_pop.future[0] == expected

        # Poisson
        n_rep = 1000
        n_list = []
        for i in range(n_rep):
            to_pop.reset()
            test_multiplier.update_data()
            n_list.append(to_pop.future[0])
        assert np.abs(np.mean(n_list) - expected) < EPS
        assert np.abs(np.std(n_list) - np.sqrt(expected)) < EPS

        # Negative binomial
        p_nb = 0.2
        nbinom_par = Parameter('nb', p_nb)
        test_multiplier.set_distribution('nbinom', nbinom_par)

        n_rep = 1000
        n_list = []
        for i in range(n_rep):
            to_pop.reset()
            test_multiplier.update_data()
            n_list.append(to_pop.future[0])
        assert np.abs(np.mean(n_list) - expected) < EPS
        assert np.abs(np.std(n_list) - np.sqrt(expected / p_nb)) < EPS


def test_class_Propagator():
    """tests to ensure the behaviour class Propagator"""
    test_model = Model('test_model')
    start = 100000
    from_pop = Population('from_pop', 0)
    from_pop.future = [start]
    to_pop = Population('to_pop', 0.)
    frac = 0.4
    fraction = Parameter('frac', frac)
    mean = 10.
    std_dev = 4.
    delay_pars = {
        'mean': Parameter('mean', mean, parameter_min=-100., parameter_max=100.),
        'sigma': Parameter('sigma', std_dev, parameter_min=-100., parameter_max=100.)
    }
    test_delay = Delay('test_delay', 'norm', delay_parameters=delay_pars, model=test_model)

    test_prop = Propagator('test_prop', from_pop, to_pop, fraction, test_delay)
    test_model.add_connector(test_prop)
    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        for func in [test_prop.update_expectation, test_prop.update_data]:
            EPS = 0.01
            if func == test_prop.update_data:
                EPS = 0.1
            to_pop.reset()
            func()
            distribution = to_pop.future
            sum_p = 0.
            sum_tp = 0.
            sum_ttp = 0.
            for i in range(len(distribution)):
                sum_p += distribution[i]
                sum_tp += i * distribution[i] * time_step
                sum_ttp += i * i * distribution[i] * time_step ** 2
            assert np.abs(sum_p - start * frac) < EPS * start * frac
            est_mean = sum_tp / (start * frac)
            assert np.abs(est_mean - mean) < EPS * mean
            est_sigma = np.sqrt(sum_ttp / (start * frac) - est_mean ** 2)
            assert np.abs(est_sigma - std_dev) < EPS * std_dev


def test_class_Splitter():
    """tests to ensure the behaviour class Splitter"""
    test_model = Model('test_model')
    start = 100000
    from_pop = Population('from_pop', 0)
    from_pop.future = [start]
    to_pops = [Population('to_pop1', 0.), Population('to_pop2', 0.)]
    fracs = [0.4, 0.6]
    fraction = Parameter('frac', fracs[0])
    mean = 10.
    std_dev = 4.
    delay_pars = {
        'mean': Parameter('mean', mean, parameter_min=-100., parameter_max=100.),
        'sigma': Parameter('sigma', std_dev, parameter_min=-100., parameter_max=100.)
    }
    test_delay = Delay('test_delay', 'norm', delay_parameters=delay_pars, model=test_model)

    test_split = Splitter('test_prop', from_pop, to_pops, [fraction], test_delay)
    test_model.add_connector(test_split)
    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        for func in [test_split.update_expectation, test_split.update_data]:
            EPS = 0.01
            if func == test_split.update_data:
                EPS = 0.1
            to_pops[0].reset()
            to_pops[1].reset()
            func()
            total = 0
            for i in range(2):
                distribution = to_pops[i].future
                frac = fracs[i]
                sum_p = 0.
                sum_tp = 0.
                sum_ttp = 0.
                for i in range(len(distribution)):
                    sum_p += distribution[i]
                    sum_tp += i * distribution[i] * time_step
                    sum_ttp += i * i * distribution[i] * time_step ** 2
                total += sum_p
                assert np.abs(sum_p - start * frac) < EPS * start * frac
                est_mean = sum_tp / (start * frac)
                assert np.abs(est_mean - mean) < EPS * mean
                est_sigma = np.sqrt(sum_ttp / (start * frac) - est_mean ** 2)
                assert np.abs(est_sigma - std_dev) < EPS * std_dev
            assert np.abs(total - start) < 0.1


def test_class_Chain():
    """tests to ensure the behaviour class Chain"""
    test_model = Model('test_model')
    from_pop = Population('from_pop', 0.)
    to_pop = Population('to_pop', 0.)
    pop_a = Population('pop_a', 0.)
    pop_b = Population('pop_b', 0.)
    start = 100000
    from_pop.future = [start]

    mean = 10.
    sigma = 2.
    delay_pars = {'mean': Parameter('mean', mean, parameter_min=0.1, parameter_max=100.),
                  'sigma': Parameter('sigma', sigma, parameter_min=0.1, parameter_max=1000.)}
    delay = Delay('delay', 'norm', delay_pars, test_model)
    frac = 0.8
    fraction = Parameter('frac', frac)
    chain = []
    chain.append(Propagator('prop_0', from_pop, pop_a, fraction, delay))
    chain.append(Propagator('prop_1', pop_a, pop_b, fraction, delay))

    test_chain = Chain('test_chain', from_pop, to_pop, chain, fraction, delay, test_model)
    test_model.add_connector(test_chain)

    for time_step in [1., 1. / 4.]:
        test_model.set_time_step(time_step)
        for func in [test_chain.update_expectation, test_chain.update_data]:
            EPS = 0.02
            if func == test_chain.update_data:
                EPS = 0.2
            to_pop.reset()
            pop_a.reset()
            pop_b.reset()
            func()
            for pop in [to_pop, pop_b]:
                distribution = pop.future
                total = start * (1. - frac ** 2) * frac
                ave = mean
                std_dev = sigma
                if pop == pop_b:
                    total = start * frac ** 2
                    ave = 2. * mean
                    std_dev = np.sqrt(2.) * sigma
                sum_p = 0.
                sum_tp = 0.
                sum_ttp = 0.
                for i in range(len(distribution)):
                    sum_p += distribution[i]
                    sum_tp += i * distribution[i] * time_step
                    sum_ttp += i * i * distribution[i] * time_step ** 2
                assert np.abs(sum_p - total) < EPS * total
                est_mean = sum_tp / total
                assert np.abs(est_mean - ave) < EPS * ave
                est_sigma = np.sqrt(sum_ttp / total - est_mean ** 2)
                assert np.abs(est_sigma - std_dev) < EPS * std_dev


def test_class_Subtractor():
    """tests to ensure the behaviour class Subtractor"""
    from_init = 20
    to_init = 10
    from_pop = Population('from_pop', from_init)
    to_pop = Population('to_pop', to_init)
    to_next = 40
    to_pop.future = [to_next]
    test_subtractor = Subtractor('test_sub', from_pop, to_pop)
    test_subtractor.update_expectation()
    assert from_pop.future[0] == -1 * to_next



# -*- coding: utf-8 -*-
"""
Delay: A class that holds information about delayed propagation

@author: karlen
"""

from scipy import stats

from pypmca.Parameter import Parameter


class Delay:
    """
    A class to define delays in propagation
        - delay_name: string, short descriptor
        - delay_type: string:
            - fast: all passed in next time step
            - norm: delay_parameters are mean and standard deviation
            - uniform: delay_parameters are mean and half width
            - erlang: delay_parameters are mean and 'k' (number of ODE stages)
        - delay_parameters: dictionary of Parameter objects. The units are days.
            - Can be None if delay_type is 'fast'
        - model: Model object that this delay is to be used with
            - Can be None if delay_type is 'fast', but otherwise must be
              specified to access the model time_step
    """

    DELAY_TYPES = ['fast', 'norm', 'uniform', 'erlang']
    DELAY_PAR_KEYS = {'fast': [], 'norm': ['mean', 'sigma'],
                      'uniform': ['mean', 'half_width'],
                      'erlang': ['mean', 'k']}
    EPSILON = 0.0001  # futures assigned until cdf reaches (1.-EPS)

    def __init__(self, delay_name: str, delay_type: str,
                 delay_parameters: dict = None, model = None):
        """Constructor
        """

        self.name = str(delay_name)
        self.delay_type = None
        self.delay_parameters = None
        self.future_expectations = None

        if delay_type != 'fast':
            if model is None:
                raise TypeError('Delay (' + self.name + ') model cannot be None')
            if not hasattr(model, 'get_time_step'):
                raise TypeError('Delay (' + self.name +
                                ') model must be a model object')
        self.model = model

        self.setup_delay(delay_type, delay_parameters)

    def __str__(self):
        return self.name

    def setup_delay(self, delay_type, delay_parameters):

        if str(delay_type) not in self.DELAY_TYPES:
            buff = '/'.join(self.DELAY_TYPES)
            raise ValueError('Delay (' + self.name + ') type "' + str(delay_type) +
                             '" invalid: not in: ' + buff)

        if delay_type != 'fast':
            if delay_parameters is None:
                raise TypeError('Delay (' + self.name + ') parameters cannot be None')
            if len(delay_parameters) < 2:
                raise ValueError('Delay (' + self.name +
                                 ') parameters dictionary must have length > 1')
            for key in delay_parameters:
                if not isinstance(delay_parameters[key], Parameter):
                    raise TypeError('Delay (' + self.name +
                                    ') parameters must be Parameter objects')
                if key == 'k':
                    if delay_parameters[key].parameter_type != 'int':
                        raise TypeError('Delay (' + self.name +
                                        ') k erlang parameter must be an integer parameter')
                elif delay_parameters[key].parameter_type != 'float':
                    raise TypeError('Delay (' + self.name +
                                    ') parameter ' + key + ' must be a float parameter')
                # flag the parameter that if changed, the parent
                # update method must be called
                delay_parameters[key].set_must_update(self)
            keys = self.DELAY_PAR_KEYS[delay_type]
            for key in keys:
                if key not in delay_parameters:
                    raise ValueError('Delay (' + self.name +
                                     ') parameters missing parameter: ' + key)

        self.delay_type = delay_type
        self.delay_parameters = delay_parameters

        self.__set_future_expectations()

    def update(self):
        self.__set_future_expectations()

    def __set_future_expectations(self):
        """
        initializes the probability for future time steps
        this needs to be redone when delay parameters are changed
        or if the model time_step changes
        """
        if self.delay_type == 'fast':
            self.future_expectations = [1.]

        else:
            time_step = self.model.get_time_step()
            day = 1. / time_step
            mean = self.delay_parameters['mean'].get_value()
            loc = 0.
            scale = 0.
            a = 0
            dist = None
            if self.delay_type == 'norm':
                loc = mean * day
                sigma = self.delay_parameters['sigma'].get_value()
                scale = sigma * day
                dist = stats.norm
            if self.delay_type == 'uniform':
                half_width = self.delay_parameters['half_width'].get_value()
                loc = mean * day - half_width * day
                scale = 2. * half_width * day
                dist = stats.uniform
            if self.delay_type == 'erlang':
                # this mimics an ODE type delay with # of stages = a 
                loc = 0.
                a = self.delay_parameters['k'].get_value()
                scale = mean * day / a
                dist = stats.erlang

            # use cdf with 0.5 time_step offsets
            cdf = 0.
            if self.delay_type != 'erlang':
                cdf = dist.cdf(0.5, loc=loc, scale=scale)
            else:
                cdf = dist.cdf(0.5, a, loc=loc, scale=scale)
            future = [cdf]
            step = 0
            while cdf < (1 - self.EPSILON):
                last_cdf = cdf
                step += 1
                if self.delay_type != 'erlang':
                    cdf = dist.cdf(step + 0.5, loc=loc, scale=scale)
                else:
                    cdf = dist.cdf(step + 0.5, a, loc=loc, scale=scale)
                increment = cdf - last_cdf
                future.append(increment)
            missing = 1. - cdf
            future[-1] += missing
            self.future_expectations = future



# -*- coding: utf-8 -*-
"""
Show the details of a pyPM.ca model by a set of tables for:
    - connectors
    - populations
    - parameters (fixed and variable)
    - variable parameters - not yet done
    - transitions: modifiers and injectors separately
    
    - The optional parameter reveal (default False), if set True will
    show all elements, even those that are "hidden" or "disabled"

@author: karlen
"""
from texttable import Texttable

from pypmca import Model, Parameter


def print_all(model):
    print(connector_table(model))
    print('\n')
    print(population_table(model))
    print('\n')
    print(parameter_table(model))
    print('\n')
    print(modifier_table(model))
    print('\n')
    print(injector_table(model))


def table_setup(model, width):
    if not isinstance(model, Model):
        raise TypeError('Error in model_table. ' +
                        ': model argument must be a Model object')

    # show table of modifiers

    table = Texttable()
    table.set_deco(Texttable.HEADER | Texttable.HLINES)
    table.set_max_width(width)
    return table


def injector_table(model, width=120, reveal=False):
    table = table_setup(model, width)

    #   transition_name, time_spec, transition_time,
    #                to_population, injection

    header = ['Injector', 'Time parameter', 'Time', 'spec', 'to (Population)',
              'Injection parameter', 'value']
    table.set_cols_dtype(['t', 't', 'a', 't', 't', 't', 'a'])

    table.set_header_align(['l', 'l', 'l', 'l', 'l', 'l', 'l'])

    rows = [header]
    for key in model.transitions:
        trans = model.transitions[key]
        if reveal or trans.enabled:
            if type(trans).__name__ == 'Injector':
                row = [str(trans), str(trans.transition_time),
                       trans.transition_time.get_value(), trans.time_spec,
                       str(trans.to_population),
                       str(trans.injection), trans.injection.get_value()]
                rows.append(row)

    table.add_rows(rows)
    return table.draw()


def modifier_table(model, width=120):
    table = table_setup(model, width)

    header = ['Modifier', 'Time parameter', 'Time', 'spec', 'Parameter',
              'before parameter', 'after parameter']
    table.set_cols_dtype(['t', 't', 'a', 't', 't', 't', 't'])

    table.set_header_align(['l', 'l', 'l', 'l', 'l', 'l', 'l'])

    rows = [header]
    for key in model.transitions:
        trans = model.transitions[key]
        if type(trans).__name__ == 'Modifier':
            row = [str(trans), str(trans.transition_time),
                   trans.transition_time.get_value(), trans.time_spec,
                   str(trans.parameter),
                   str(trans.parameter_before),
                   str(trans.parameter_after)]
            rows.append(row)

    table.add_rows(rows)
    return table.draw()


def get_connector_row(connector):
    con = connector
    row = [str(con), type(con).__name__]
    for pop in [con.from_population, con.to_population]:
        buff = []
        if isinstance(pop, list):
            for item in pop:
                buff.append(str(item))
        else:
            buff.append(str(pop))
        row.append('\n'.join(buff))
    names = []
    types = []
    values = []
    status = []
    for par_key in con.parameters:
        par = con.parameters[par_key]
        names.append(str(par))
        types.append(par.parameter_type)
        if par.parameter_type in ['int', 'bool']:
            values.append(str(par.get_value()))
        else:
            values.append("{:10.3f}".format(par.get_value()))
        status.append(par.get_status())
    row.append('\n'.join(names))
    row.append('\n'.join(values))
    return row


def connector_table(model, width=120):
    table = table_setup(model, width)

    header = ['Connector', 'Type', 'From\n(populations)', 'To\n(populations)',
              'Parameters', 'Values']
    table.set_cols_dtype(['t', 't', 't', 't',
                          't', 't'])
    table.set_header_align(['l', 'l', 'l', 'l',
                            'l', 'r'])

    rows = [header]
    # order is important!
    for key in model.connector_list:
        con = model.connectors[key]
        rows.append(get_connector_row(con))
        if type(con).__name__ == 'Chain':
            for chain_con in con.chain:
                row = get_connector_row(chain_con)
                row[0] = str(con) + ': ' + row[0]
                row[1] = '->' + row[1]
                rows.append(row)

    table.add_rows(rows)
    return table.draw()


def population_table(model, width=120):
    table = table_setup(model, width)

    header = ['Population', 'Description', 'Parameter', 'Start value']
    table.set_cols_dtype(['t', 't', 't', 'a'])
    table.set_header_align(['l', 'l', 'l', 'l'])

    rows = [header]

    for key in model.populations:
        pop = model.populations[key]
        row = [str(pop), pop.description]
        init = pop.initial_value
        if isinstance(init, Parameter):
            row.append(str(init))
            row.append(init.get_value())
        else:
            row.append(' ')
            row.append(init)

        rows.append(row)

    table.add_rows(rows)

    return table.draw()


def get_par_row(par):
    buff = []
    buff.append(str(par))
    buff.append(par.description)
    buff.append(par.get_min())
    buff.append(par.get_max())
    buff.append(par.initial_value)
    buff.append(par.get_value())
    return buff


def parameter_table(model, width=120):
    table = table_setup(model, width)

    header = ['Parameter', 'Description', 'min', 'max', 'Init', 'Value']
    table.set_cols_dtype(['t', 't', 'a', 'a', 'a', 'a'])
    table.set_header_align(['l', 'l', 'l', 'l', 'l', 'l'])

    par_dict = {}

    for key in model.connectors:
        con = model.connectors[key]
        for par_key in con.parameters:
            par = con.parameters[par_key]
            row = get_par_row(par)
            if len(row) > 0:
                par_dict[row[0]] = row
        if type(con).__name__ == 'Chain':
            for chain_con in con.chain:
                for par_key in chain_con.parameters:
                    par = chain_con.parameters[par_key]
                    row = get_par_row(par)
                    if len(row) > 0:
                        par_dict[row[0]] = row

    for key in model.populations:
        pop = model.populations[key]
        init = pop.initial_value
        if isinstance(init, Parameter):
            row = get_par_row(init)
            par_dict[row[0]] = row

    for key in model.transitions:
        trans = model.transitions[key]
        for par_key in trans.parameters:
            par = trans.parameters[par_key]
            row = get_par_row(par)
            if len(row) > 0:
                par_dict[row[0]] = row

    par_list = list(par_dict.keys())
    par_list.sort()

    rows = [header]
    for item in par_list:
        rows.append(par_dict[item])

    table.add_rows(rows)

    return table.draw()


def get_var_par_row(par):
    buff = []
    buff.append(str(par))
    buff.append(par.description)
    buff.append(par.get_min())
    buff.append(par.get_max())
    buff.append(par.prior_function)
    buff.append(par.prior_parameters['mean'])
    if par.prior_function == 'uniform':
        buff.append(par.prior_parameters['half_width'])
    else:
        buff.append(par.prior_parameters['sigma'])
    buff.append(par.mcmc_step)
    buff.append(par.initial_value)
    buff.append(par.get_value())
    return buff


def variable_parameter_table(model, width=120):
    table = table_setup(model, width)

    header = ['Parameter', 'Description', 'min', 'max', 'prior', 'prior_mean', 'prior_w', 'mcmc_step', 'Init', 'Value']
    table.set_cols_dtype(['t', 't', 'a', 'a', 't', 'e', 'e', 'e', 'a', 'a'])
    table.set_header_align(['l', 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l'])

    par_dict = {}

    for key in model.connectors:
        con = model.connectors[key]
        for par_key in con.parameters:
            par = con.parameters[par_key]
            if par.get_status() == 'variable':
                row = get_var_par_row(par)
                if len(row) > 0:
                    par_dict[row[0]] = row
        if type(con).__name__ == 'Chain':
            for chain_con in con.chain:
                for par_key in chain_con.parameters:
                    par = chain_con.parameters[par_key]
                    if par.get_status() == 'variable':
                        row = get_var_par_row(par)
                        if len(row) > 0:
                            par_dict[row[0]] = row

    for key in model.populations:
        pop = model.populations[key]
        init = pop.initial_value
        if isinstance(init, Parameter):
            if init.get_status() == 'variable':
                row = get_var_par_row(init)
                par_dict[row[0]] = row

    for key in model.transitions:
        trans = model.transitions[key]
        for par_key in trans.parameters:
            par = trans.parameters[par_key]
            if par.get_status() == 'variable':
                row = get_var_par_row(par)
                if len(row) > 0:
                    par_dict[row[0]] = row

    par_list = list(par_dict.keys())
    par_list.sort()

    rows = [header]
    for item in par_list:
        rows.append(par_dict[item])

    table.add_rows(rows)

    return table.draw()


def get_delay_row(connector, delay):
    con = connector
    row = [str(con), type(con).__name__]
    row.append(str(delay))
    row.append(delay.delay_type)
    if delay.delay_type != 'fast':
        delay_parameters = delay.delay_parameters
        mean_par = delay_parameters['mean']
        width_par = None
        if delay.delay_type == 'norm':
            width_par = delay_parameters['sigma']
        elif delay.delay_type == 'uniform':
            width_par = delay_parameters['half_width']
        elif delay.delay_type == 'erlang':
            width_par = delay_parameters['k']

        row.append(str(mean_par))
        row.append(mean_par.get_value())
        row.append(str(width_par))
        row.append(width_par.get_value())
    else:
        row.append('')
        row.append('')
        row.append('')
        row.append('')
    return row


def delay_table(model, width=120):
    table = table_setup(model, width)

    header = ['Connector', 'Con Type', 'Delay name', 'Delay Type',
              'mean par', 'mean val', 'width par', 'width val']
    table.set_cols_dtype(['t', 't', 't', 't', 't',
                          'a', 't', 'a'])
    table.set_header_align(['l', 'l', 'l', 'l', 'l',
                            'l', 'l', 'l'])

    rows = [header]
    for key in model.connectors:
        con = model.connectors[key]
        if hasattr(con, 'delay'):
            if isinstance(con.delay, list):
                for delay in con.delay:
                    rows.append(get_delay_row(con, delay))
                    if type(con).__name__ == 'Chain':
                        for chain_con in con.chain:
                            row = get_delay_row(chain_con, chain_con.delay)
                            row[0] = str(con) + ': ' + row[0]
                            row[1] = '->' + row[1]
                            rows.append(row)

            else:
                rows.append(get_delay_row(con, con.delay))
                if type(con).__name__ == 'Chain':
                    for chain_con in con.chain:
                        row = get_delay_row(chain_con, chain_con.delay)
                        row[0] = str(con) + ': ' + row[0]
                        row[1] = '->' + row[1]
                        rows.append(row)

    table.add_rows(rows)
    return table.draw()



# -*- coding: utf-8 -*-
"""
Connector: A base'virtual' class for defining connections between populations.

The derived classes must define the methods for updating the future values of populations.

Depending on the connector type, from_population and to_population can be a
single population or a list of populations.

@author:karlen
"""

from pypmca.Population import Population


class Connector:
    """
    Base class for defining connections between populations
    """

    def __init__(self, connector_name, from_population, to_population):
        """Constructor
        """
        if connector_name.find(',') > -1:
            raise ValueError('Error in constructing ' + self.name +
                             ': name cannot contain a comma.')
        self.name = str(connector_name)

        pops = {'from_population': from_population,
                'to_population': to_population}
        for key in pops:
            if isinstance(pops[key], list):
                for pop in pops[key]:
                    if not isinstance(pop, Population):
                        raise TypeError('Error in constructing ' + self.name +
                                        ': ' + key + ' list must contain Population objects')
            else:
                if not isinstance(pops[key], Population):
                    raise TypeError('Error in constructing ' + self.name +
                                    ': ' + key + ' argument must be a Population object')

        match = False
        if isinstance(from_population, list):
            for fpop in from_population:
                if isinstance(to_population, list):
                    for tpop in to_population:
                        if fpop == tpop:
                            match = True
                else:
                    if fpop == to_population:
                        match = True
        else:
            if isinstance(to_population, list):
                for tpop in to_population:
                    if from_population == tpop:
                        match = True
            else:
                if from_population == to_population:
                    match = True

        if match:
            raise ValueError('Error in constructing ' + self.name +
                             ': from_ and to_ populations cannot be the same')

        self.from_population = from_population
        self.to_population = to_population
        self.parameters = {}

    def __str__(self):
        return self.name

    def update_expectation(self):
        """
        This method must be defined in derived classes

        Raises
        ------
        NotImplementedError if not defined in the derived class

        Returns
        -------
        None.

        """
        raise NotImplementedError()

    def update_data(self):
        """
        This method must be defined in derived classes

        Raises
        ------
        NotImplementedError if not defined in the derived class

        Returns
        -------
        None.

        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
Multiplier: A Connector class that propagates a population determined by
the product of the two "from_populations" to the "to_population" either in
the next time step or distributed in time:

@author: karlen
"""

from scipy import stats

from pypmca.Connector import Connector
from pypmca.Delay import Delay
from pypmca.Parameter import Parameter


class Multiplier(Connector):
    """
    Multiplier send a population determined by the product of
    two "from_populations" to the "to_population",
    either in the next time step or distributed in time:
        - connector_name: string, short descriptor
        - from_population: list of two Population objects to be multiplied
        If a third from_population is included, it enters in the denominator
        - to_population: Population object that is the destination population
        - scale_parameter: Parameter object that multiplies the product
        of the from_population totals
        - delay: Delay object that defines how propagation is spread over time.
        - model: necessary to access the time_step
        - distribution: newly infected production distribution. 
            poisson (indendent infections)
            nbinom (negative binomial - in popular use)
            nbinom_par = p in the scipy nbinom. limit 0.001<p<0.999
    """

    def __init__(self, connector_name: str, from_population: list, to_population,
                 scale_parameter: Parameter, delay: Delay, model=None, distribution: str = 'poisson',
                 nbinom_par: Parameter = None):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, list):
            raise TypeError('Multiplier (' + self.name +
                            ') from_population must be a list')

        if len(from_population) not in [2, 3]:
            raise ValueError('Multiplier (' + self.name +
                             ') from_population list length must be 2 or 3')

        if not isinstance(scale_parameter, Parameter):
            raise TypeError('Multiplier (' + self.name +
                            ') scale_parameter must be a Parameter object')
        self.scale_parameter = scale_parameter

        if not isinstance(delay, Delay):
            raise TypeError('Multiplier (' + self.name +
                            ') delay must be a Delay object')
        self.delay = delay

        if model is None:
            raise TypeError('Multiplier (' + self.name + ') model cannot be None')
        if not hasattr(model, 'get_time_step'):
            raise TypeError('Multiplier (' + self.name +
                            ') model must be a model object')
        self.model = model

        self.parameters[str(self.scale_parameter)] = self.scale_parameter
        if delay.delay_parameters is not None:
            for key in delay.delay_parameters:
                self.parameters['delay_' + key] = delay.delay_parameters[key]

        self.__distribution = None
        self.__nbinom_par = None
        self.set_distribution(distribution, nbinom_par)

    def set_distribution(self, distribution, nbinom_par):
        if distribution not in ['poisson', 'nbinom']:
            raise ValueError('Multiplier (' + self.name +
                             ') distribution must be poisson or nbinom')
        self.__distribution = distribution

        if distribution == 'nbinom' and \
                (nbinom_par is None or not isinstance(nbinom_par, Parameter)):
            raise TypeError('Multiplier (' + self.name +
                            ') nbinom_par must be a Parameter object')
        self.__nbinom_par = nbinom_par

        if distribution == 'nbinom':
            self.parameters['nbinom_par'] = self.__nbinom_par
        # in case parameter changed, update the model list of parameters
        self.model.update_lists()

    def get_distribution(self):
        return self.__distribution, self.__nbinom_par

    def update_expectation(self):
        """
        Calculate contributions to other populations by updating their
        future_expectations
        """
        scale = self.__get_scale()
        self.to_population.update_future_expectation(scale, self.delay)

    def update_data(self):
        """
        Simulate data for other populations by updating their
        future_expectations
        """
        scale = self.__get_scale()
        if self.__distribution == 'poisson':
            n = stats.poisson.rvs(scale)
            self.to_population.update_future_data(n, self.delay)
        else:
            p = self.__nbinom_par.get_value()
            if p < 0.001:
                p = 0.001
            if p > 0.999:
                p = 0.999
            r = scale * p / (1. - p)
            n = stats.nbinom.rvs(r, p)
            self.to_population.update_future_data(n, self.delay)

    def __get_scale(self):
        """
        Calculate expected number to be sent to "to_population"
        """
        denom = 1.
        if len(self.from_population) == 3 and \
                self.from_population[2].history[-1] > 0:
            denom = self.from_population[2].history[-1]

        ratio = 1. * self.from_population[0].history[-1] / denom
        scale = self.scale_parameter.get_value() * ratio * self.from_population[1].history[-1]
        scale *= self.model.get_time_step()
        return scale



# -*- coding: utf-8 -*-
"""
Modifier: A Transition class that allows a parameter value to be modified at
a specific point in the evolution

@author:karlen
"""

from pypmca.Parameter import Parameter
from pypmca.Transition import Transition


class Modifier(Transition):
    """
    A modifier object will cause a parameter value to change at specified point
    in the evolution.

    The time_step for which the action is taken can be specified as one of 'time_spec':
    - 'rel_days': int, number of days after start (t0)
    - 'rel_steps': int, number of time_steps after start (t0)
    
    After the time_step is reached in the model, the value for parameter is
    changed to the value contained in parameter_after.
    
    The _before value of the parameter is specified so that
    the transition can be reversed. To restart the model the transitions are
    reset in reverse order, so that the _before value for the first transition
    is used at the beginning.

    The transition time is specified by a Parameter object.

    Transitions are not applied during the model "boot"

    """

    def __init__(self, transition_name: str, time_spec: str, transition_time: Parameter, parameter: Parameter,
                 parameter_before: Parameter, parameter_after: Parameter, enabled: bool = True, model=None):
        """Constructor
        """
        super().__init__(transition_name, time_spec, transition_time, enabled, model)

        if parameter is None:
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter_before cannot be None')

        if parameter_before is None:
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter_before cannot be None')

        if parameter_after is None:
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter_after cannot be None')

        if not isinstance(parameter, Parameter):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter must be a Parameter object')

        if not isinstance(parameter_before, Parameter):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter_before must be a Parameter object')

        if not isinstance(parameter_after, Parameter):
            raise TypeError('Error in constructing transition (' + self.name +
                            '): parameter_after must be a Parameter object')

        self.parameter = parameter
        self.parameter_before = parameter_before
        self.parameter_after = parameter_after

        self.initial_value = parameter_before.get_value()

        self.parameters[str(parameter_after)] = parameter_after
        self.parameters[str(parameter_before)] = parameter_before

    def take_action(self, expectations=True):
        """
        Modify the value of the parameter
        """
        self.parameter.set_value(self.parameter_after.get_value())

    def reset(self):
        """
        Revert to the value of the before parameter
        """
        self.parameter.set_value(self.parameter_before.get_value())



#!/usr/bin/env python

"""Tests for `pypmca` package."""

import pytest

from pypmca import Model, Population, Delay, Parameter, Multiplier, Propagator, \
    Splitter, Adder, Subtractor, Chain, Modifier, Injector, Ensemble
import numpy as np
import copy
from pathlib import Path

example_dir = Path('../../examples/').resolve()
path_model_2 = example_dir / 'ref_model_2.pypm'
path_model_1 = example_dir / 'ref_model_1.pypm'

def test_Model_properties():
    """tests to ensure the properties of Model"""
    ref_model_2 = Model.open_file(path_model_2)
    ref_model_1 = Model.open_file(path_model_1)

    for test_model in [ref_model_1, ref_model_2]:
        # check simple scaling of initial contagious population
        # Will not be exact due to bootstrap
        EPS = 0.01
        cont_0 = test_model.parameters['cont_0'].get_value()
        test_model.reset()
        test_model.evolve_expectations(200)
        n_reported1 = test_model.populations['reported'].history[-1]
        assert test_model.populations['contagious'].history[-1] < 10.

        test_model.parameters['cont_0'].set_value(2 * cont_0)
        test_model.reset()
        test_model.evolve_expectations(200)
        n_reported2 = test_model.populations['reported'].history[-1]
        assert np.abs(n_reported2 - n_reported1 * 2) < EPS * n_reported1 * 2

    # check that the mean of many data runs is near expectation

    ref_model = Model.open_file(path_model_2)
    ref_model.reset()
    n_days = 60
    ref_model.evolve_expectations(n_days)
    sim_models = []
    n_rep = 100
    for i in range(n_rep):
        sim_model = copy.deepcopy(ref_model)
        sim_model.reset()
        sim_model.generate_data(n_days)
        sim_models.append(sim_model)
    for pop_name in ref_model.populations:
        pop = ref_model.populations[pop_name]
        if pop.show_sim:
            results = []
            for sim_model in sim_models:
                results.append(sim_model.populations[pop_name].history[-1])
            mean = np.mean(np.array(results))
            std = np.std(np.array(results))
            error = std / np.sqrt(1. * n_rep)
            expect = ref_model.populations[pop_name].history[-1]
            assert np.abs(expect - mean) < 8. * error


def test_Ensemble_properties_identical():
    """tests to ensure the properties of Ensemble"""
    # Test that the ensemble of two identical models
    # behalves like twice a single model.
    # independent of the contact matrix
    test_a = Model.open_file(path_model_2)
    test_a.name = 'test_a'
    test_b = Model.open_file(path_model_2)
    test_b.name = 'test_b'
    reference = Model.open_file(path_model_2)
    reference.name = 'reference'
    single = Model.open_file(path_model_2)
    single.name = 'single'

    test_ensemble = Ensemble('test_ensemble', reference)
    test_ensemble.upload_models([test_a, test_b])

    contacts = [
        [[1., 0.0], [0.0, 1.]],
        [[1., 1.0], [1.0, 1.]],
        [[1., 0.5], [0.7, 1.]],
    ]

    for contact in contacts:
        test_ensemble.define_cross_transmission('infection cycle', 'infected',
                                                'susceptible', 'total',
                                                'contagious', 'alpha',
                                                contact_type='fixed',
                                                contact=contact)

        n_days = 100
        test_ensemble.reset()
        test_ensemble.evolve_expectations(n_days)

        single.reset()
        single.evolve_expectations(n_days)

        for pop_name in test_ensemble.populations:
            pop = test_ensemble.populations[pop_name]
            if pop.show_sim:
                ens_hist = test_ensemble.populations[pop_name].history
                single_hist = single.populations[pop_name].history
                for i in range(len(ens_hist)):
                    ratio = ens_hist[i] / single_hist[i]
                    assert np.abs(ratio - 2.) < 0.001


def test_Ensemble_properties_different():
    """tests to ensure the properties of Ensemble with different sub models"""
    # Test that the ensemble of two identical models
    # behalves like twice a single model.
    # Only for the independent (ie. diagonal) case
    # Note: this test would fail if a fixed contact_matrix
    # is passed which happens to be the identity matrix
    # The difference is that when specified as independent
    # each model is booted independently
    # If a contact matrix is specified, then the boot goal
    # is the combined total of all models. So there will be a different starting point.
    test_a = Model.open_file(path_model_2)
    test_a.name = 'test_a'
    test_b = Model.open_file(path_model_2)
    test_b.name = 'test_b'
    test_b.parameters['alpha_0'].set_value(0.7)
    test_c = Model.open_file(path_model_2)
    test_c.name = 'test_c'
    test_d = Model.open_file(path_model_2)
    test_d.name = 'test_d'
    test_d.parameters['alpha_0'].set_value(0.7)
    reference = Model.open_file(path_model_2)
    reference.name = 'reference'

    test_ensemble = Ensemble('test_ensemble', reference)
    test_ensemble.upload_models([test_a, test_b])
    test_ensemble.define_cross_transmission('infection cycle', 'infected',
                                            'susceptible', 'total',
                                            'contagious', 'alpha',
                                            contact_type='diagonal')

    n_days = 100
    test_c.reset()
    test_c.evolve_expectations(n_days)
    test_d.reset()
    test_d.evolve_expectations(n_days)
    test_ensemble.reset()
    test_ensemble.evolve_expectations(n_days)
    for pop_name in test_ensemble.populations:
        pop = test_ensemble.populations[pop_name]
        if pop.show_sim:
            ens_hist = test_ensemble.populations[pop_name].history
            tc_hist = test_c.populations[pop_name].history
            td_hist = test_d.populations[pop_name].history
            for i in range(len(ens_hist)):
                ratio = ens_hist[i] / (tc_hist[i] + td_hist[i])
                assert np.abs(ratio - 1.) < 0.01



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import versioneer

descr = """
The ``pyPM.ca`` population modeller (www.pyPM.ca) describes connected systems with
discrete-time difference equations. It was developed specifically
to understand and characterize the CoViD-19 epidemic.

A ``pyPM.ca`` model is built by connecting a set of population objects with
connector objects. The connectors represent either a transfer that occurs
immediately at the next time step
or one which is delayed and distributed in time. Each
population object retains a record of its size at each time step, and also
maintains a list of future contributions, arising from delayed transfers
from other populations.
Calculations of population size are done either in terms of expectation values 
or simulated data,
allowing the model to be used for both analysis and simulation.
"""

requirements = ['scipy', 'numpy']

setup_requirements = ['pytest-runner', ] + requirements

test_requirements = ['pytest>=3', ] + requirements

setup(
        author="Dean Karlen",
        author_email='karlen@uvic.ca',
        python_requires='>=3.5',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            ],
        description="population modeller",
        install_requires=requirements,
        license="GNU General Public License v3",
        long_description=descr,
        include_package_data=True,
        keywords='pypmca',
        name='pypmca',
        version=versioneer.get_version(),
        packages=find_packages(include=['pypmca', 'pypmca.*']),
        setup_requires=setup_requirements,
        test_suite='tests',
        tests_require=test_requirements,
        url='https://github.com/pypm/pypmca',
        zip_safe=False,
        cmdclass=versioneer.get_cmdclass()
        )
